import os
import shutil
from glob import glob

import packaging.version
from qtpy.QtWidgets import QMessageBox, QWidget

from PartSegCore import state_store

from .. import parsed_version
from .._launcher.check_version import IGNORE_FILE
from . import napari_get_settings


def import_config():
    if os.path.exists(state_store.save_folder):
        return
    version = packaging.version.parse(parsed_version.base_version)
    base_folder = os.path.dirname(state_store.save_folder)
    possible_folders = glob(os.path.join(base_folder, "*"))
    versions = list(
        sorted(
            (
                x
                for x in [packaging.version.parse(os.path.basename(y)) for y in possible_folders]
                if isinstance(x, packaging.version.Version)
            ),
            reverse=True,
        )
    )
    before_version = None
    for x in versions:
        if x < version:
            before_version = x
            break
    if before_version is not None:
        before_name = str(before_version)
        widget = QWidget()
        resp = QMessageBox.question(
            widget,
            "Import from old version",
            "There is no configuration folder for this version of PartSeg\n"
            "Would you like to import it from " + before_name + " version of PartSeg",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.Yes,
        )
        if resp == QMessageBox.Yes:
            shutil.copytree(os.path.join(base_folder, before_name), state_store.save_folder)
            if os.path.exists(os.path.join(state_store.save_folder, IGNORE_FILE)):
                os.remove(os.path.join(state_store.save_folder, IGNORE_FILE))
            napari_settings = napari_get_settings(state_store.save_folder)
            if hasattr(napari_settings, "load") and napari_settings.load is not None:
                napari_settings.load()
            elif hasattr(napari_settings, "_load") and napari_settings._load is not None:  # pylint: disable=W0212
                napari_settings._load()  # pylint: disable=W0212



import importlib
import itertools
import pkgutil
import typing

import pkg_resources


def get_plugins():

    packages = pkgutil.iter_modules(__path__, __name__ + ".")
    packages2 = itertools.chain(
        pkg_resources.iter_entry_points("PartSegCore.plugins"),
        pkg_resources.iter_entry_points("partsegcore.plugins"),
    )
    return [importlib.import_module(el.name) for el in packages] + [el.load() for el in packages2]


plugins_loaded = set()


def register():
    for el in get_plugins():
        if hasattr(el, "register") and el.__name__ not in plugins_loaded:
            assert isinstance(el.register, typing.Callable)  # nosec
            el.register()
            plugins_loaded.add(el.__name__)


def register_if_need():
    if len(plugins_loaded) == 0:
        register()



import sys

import sentry_sdk
from qtpy.QtWidgets import QMessageBox
from superqt import ensure_main_thread

from PartSeg import parsed_version
from PartSegCore import state_store
from PartSegCore.segmentation.algorithm_base import SegmentationLimitException
from PartSegImage import TiffFileException


def my_excepthook(type_, value, trace_back):
    """
    Custom excepthook. base on base on :py:data:`state_store.show_error_dialog` decide if shown error dialog.

    """

    # log the exception here
    if state_store.show_error_dialog and not isinstance(value, KeyboardInterrupt):
        if state_store.report_errors and parsed_version.is_devrelease:
            with sentry_sdk.push_scope() as scope:
                scope.set_tag("auto_report", "true")
                sentry_sdk.capture_exception(value)
        try:
            show_error(value)
        except ImportError:
            sys.__excepthook__(type_, value, trace_back)
    elif isinstance(value, KeyboardInterrupt):
        print("KeyboardInterrupt close", file=sys.stderr)
        sys.exit(1)
    else:
        # then call the default handler
        sys.__excepthook__(type_, value, trace_back)


@ensure_main_thread
def show_error(error=None):
    """This class create error dialog and show it"""
    if error is None:
        return

    if isinstance(error, TiffFileException):
        mess = QMessageBox()
        mess.setIcon(QMessageBox.Critical)
        mess.setText("During read file there is an error: " + error.args[0])
        mess.setWindowTitle("Tiff error")
        mess.exec_()
        return
    if isinstance(error, SegmentationLimitException):
        mess = QMessageBox()
        mess.setIcon(QMessageBox.Critical)
        mess.setText("During segmentation process algorithm meet limitations:\n" + "\n".join(error.args))
        mess.setWindowTitle("Segmentation limitations")
        mess.exec_()
        return
    from PartSeg.common_gui.error_report import ErrorDialog

    dial = ErrorDialog(error, "Exception during program run")
    dial.exec_()


@ensure_main_thread
def show_warning(header=None, text=None):
    """show warning :py:class:`PyQt5.QtWidgets.QMessageBox`"""
    message = QMessageBox(QMessageBox.Warning, header, text, QMessageBox.Ok)
    message.exec_()



import dataclasses
import json
import os
import sys
import tarfile
import typing
import warnings
from collections import defaultdict
from contextlib import suppress
from functools import partial
from io import BufferedIOBase, BytesIO, IOBase, RawIOBase, TextIOBase
from pathlib import Path

import numpy as np
import tifffile
from napari.plugins._builtins import napari_write_points

from PartSegImage import BaseImageWriter, GenericImageReader, Image, IMAGEJImageWriter, ImageWriter, TiffImageReader
from PartSegImage.image import FRAME_THICKNESS, reduce_array

from ..algorithm_describe_base import AlgorithmProperty, Register, ROIExtractionProfile
from ..io_utils import (
    LoadBase,
    LoadPoints,
    SaveBase,
    SaveMaskAsTiff,
    SaveROIAsNumpy,
    SaveROIAsTIFF,
    SegmentationType,
    UpdateLoadedMetadataBase,
    WrongFileTypeException,
    check_segmentation_type,
    get_tarinfo,
    open_tar_file,
    proxy_callback,
    tar_to_buff,
)
from ..json_hooks import ProfileEncoder
from ..project_info import AdditionalLayerDescription, HistoryElement, ProjectInfoBase
from ..roi_info import ROIInfo

if sys.version_info[:3] == (3, 9, 7):
    ProjectInfoBase = object  # noqa: F811


@dataclasses.dataclass(frozen=True)
class MaskProjectTuple(ProjectInfoBase):
    """
    Dataclass instance to describe segmentation state

    :ivar str file_path: path to current processed file
    :ivar typing.Union[Image,str,None] ~.image: image which is proceeded in given segmentation.
        If :py:class:`str` then it is path to image on drive
    :ivar typing.Optional[np.ndarray] ~.mask: Mask limiting segmentation area.
    :ivar typing.Optional[np.ndarray] ~.roi: ROI array.
    :ivar SegmentationInfo ~.roi_info: ROI description
    :ivar typing.List[int] ~.selected_components: list of selected components
    :ivar typing.Dict[int,typing.Optional[SegmentationProfile]] ~.segmentation_parameters:
        For each component description set of parameters used for segmentation
    :ivar typing.List[HistoryElement] history: list of operations needed to create :py:attr:`mask`
    :ivar str ~.errors: information about problems meet during calculation
    :ivar typing.Optional[typing.List[float]] ~.spacing: information about spacing when image is missed.
        For napari read plugin
    """

    file_path: str
    image: typing.Union[Image, str, None]
    mask: typing.Optional[np.ndarray] = None
    roi_info: ROIInfo = ROIInfo(None)
    additional_layers: typing.Dict[str, AdditionalLayerDescription] = dataclasses.field(default_factory=dict)
    selected_components: typing.List[int] = dataclasses.field(default_factory=list)
    roi_extraction_parameters: typing.Dict[int, typing.Optional[ROIExtractionProfile]] = dataclasses.field(
        default_factory=dict
    )
    history: typing.List[HistoryElement] = dataclasses.field(default_factory=list)
    errors: str = ""
    spacing: typing.Optional[typing.List[float]] = None
    points: typing.Optional[np.ndarray] = None

    def get_raw_copy(self):
        return MaskProjectTuple(self.file_path, self.image.substitute(mask=None))

    def is_raw(self):
        return self.roi_info.roi is None

    def is_masked(self):
        return self.mask is not None

    def get_raw_mask_copy(self):
        return MaskProjectTuple(file_path=self.file_path, image=self.image.substitute(), mask=self.mask)

    @property
    def roi(self):
        warnings.warn("roi is deprecated", DeprecationWarning, 2)
        return self.roi_info.roi


def save_stack_segmentation(
    file_data: typing.Union[tarfile.TarFile, str, Path, TextIOBase, BufferedIOBase, RawIOBase, IOBase],
    segmentation_info: MaskProjectTuple,
    parameters: dict,
    range_changed=None,
    step_changed=None,
):
    if range_changed is None:
        range_changed = empty_fun
    if step_changed is None:
        step_changed = empty_fun
    range_changed(0, 7)
    tar_file, file_path = open_tar_file(file_data, "w")
    step_changed(1)
    try:
        segmentation_buff = BytesIO()
        # noinspection PyTypeChecker
        if segmentation_info.image is not None:
            spacing = segmentation_info.image.spacing
        else:
            spacing = parameters.get("spacing", (10 ** -6, 10 ** -6, 10 ** -6))
        segmentation_image = Image(
            segmentation_info.roi_info.roi, spacing, axes_order=Image.axis_order.replace("C", "")
        )
        try:
            ImageWriter.save(segmentation_image, segmentation_buff, compression=None)
        except ValueError:
            segmentation_buff.seek(0)
            tifffile.imwrite(segmentation_buff, segmentation_info.roi_info.roi)
        segmentation_tar = get_tarinfo("segmentation.tif", segmentation_buff)
        tar_file.addfile(segmentation_tar, fileobj=segmentation_buff)
        step_changed(3)
        metadata = {
            "components": [int(x) for x in segmentation_info.selected_components],
            "parameters": {str(k): v for k, v in segmentation_info.roi_extraction_parameters.items()},
            "shape": segmentation_info.roi_info.roi.shape,
            "annotations": segmentation_info.roi_info.annotations,
        }
        if isinstance(segmentation_info.image, Image):
            file_path = segmentation_info.image.file_path
        elif isinstance(segmentation_info.image, str):
            file_path = segmentation_info.image
        else:
            file_path = ""
        if file_path != "":
            if parameters["relative_path"] and isinstance(file_data, str):
                metadata["base_file"] = os.path.relpath(file_path, os.path.dirname(file_data))
            else:
                metadata["base_file"] = file_path
        metadata_buff = BytesIO(json.dumps(metadata, cls=ProfileEncoder).encode("utf-8"))
        metadata_tar = get_tarinfo("metadata.json", metadata_buff)
        tar_file.addfile(metadata_tar, metadata_buff)
        step_changed(4)
        if segmentation_info.mask is not None:
            mask = segmentation_info.mask
            if mask.dtype == bool:
                mask = mask.astype(np.uint8)
            mask_buff = BytesIO()
            tifffile.imwrite(mask_buff, mask)
            mask_tar = get_tarinfo("mask.tif", mask_buff)
            tar_file.addfile(mask_tar, fileobj=mask_buff)
        if segmentation_info.roi_info.alternative:
            alternative_buff = BytesIO()
            np.savez(alternative_buff, **segmentation_info.roi_info.alternative)
            alternative_tar = get_tarinfo("alternative.npz", alternative_buff)
            tar_file.addfile(alternative_tar, fileobj=alternative_buff)
        step_changed(5)
        el_info = []
        for i, hist in enumerate(segmentation_info.history):
            el_info.append(
                {
                    "index": i,
                    "mask_property": hist.mask_property,
                    "segmentation_parameters": hist.roi_extraction_parameters,
                    "annotations": hist.annotations,
                }
            )
            hist.arrays.seek(0)
            hist_info = get_tarinfo(f"history/arrays_{i}.npz", hist.arrays)
            hist.arrays.seek(0)
            tar_file.addfile(hist_info, hist.arrays)
        if el_info:
            hist_str = json.dumps(el_info, cls=ProfileEncoder)
            hist_buff = BytesIO(hist_str.encode("utf-8"))
            tar_algorithm = get_tarinfo("history/history.json", hist_buff)
            tar_file.addfile(tar_algorithm, hist_buff)
        step_changed(6)

    finally:
        if isinstance(file_data, (str, Path)):
            tar_file.close()
    step_changed(6)


def load_stack_segmentation(file_data: typing.Union[str, Path], range_changed=None, step_changed=None):
    if range_changed is None:
        range_changed = empty_fun
    if step_changed is None:
        step_changed = empty_fun
    range_changed(0, 7)
    tar_file = open_tar_file(file_data)[0]
    try:
        if check_segmentation_type(tar_file) != SegmentationType.mask:
            raise WrongFileTypeException()
        files = tar_file.getnames()
        step_changed(1)
        metadata = load_metadata(tar_file.extractfile("metadata.json").read().decode("utf8"))
        step_changed(2)
        if "segmentation.npy" in files:
            segmentation_file_name = "segmentation.npy"
            segmentation_load_fun = np.load
        else:
            segmentation_file_name = "segmentation.tif"
            segmentation_load_fun = TiffImageReader.read_image
        segmentation_buff = BytesIO()
        segmentation_tar = tar_file.extractfile(tar_file.getmember(segmentation_file_name))
        segmentation_buff.write(segmentation_tar.read())
        step_changed(3)
        segmentation_buff.seek(0)
        roi = segmentation_load_fun(segmentation_buff)
        if isinstance(roi, Image):
            spacing = roi.spacing
            roi = roi.get_channel(0)
        else:
            spacing = None
        step_changed(4)
        if "mask.tif" in tar_file.getnames():
            mask = tifffile.imread(tar_to_buff(tar_file, "mask.tif"))
            if np.max(mask) == 1:
                mask = mask.astype(bool)
        else:
            mask = None
        if "alternative.npz" in tar_file.getnames():
            alternative = np.load(tar_to_buff(tar_file, "alternative.npz"))
        else:
            alternative = {}
        roi_info = ROIInfo(reduce_array(roi), annotations=metadata.get("annotations", {}), alternative=alternative)
        step_changed(5)
        history = []
        with suppress(KeyError):
            history_buff = tar_file.extractfile(tar_file.getmember("history/history.json")).read()
            history_json = load_metadata(history_buff)
            for el in history_json:
                history_buffer = BytesIO()
                history_buffer.write(tar_file.extractfile(f"history/arrays_{el['index']}.npz").read())
                history_buffer.seek(0)
                history.append(
                    HistoryElement(
                        roi_extraction_parameters=el["segmentation_parameters"],
                        mask_property=el["mask_property"],
                        arrays=history_buffer,
                        annotations=el.get("annotations", {}),
                    )
                )
        step_changed(6)
    finally:
        if isinstance(file_data, (str, Path)):
            tar_file.close()
    return MaskProjectTuple(
        file_path=file_data if isinstance(file_data, str) else "",
        image=metadata["base_file"] if "base_file" in metadata else None,
        roi_info=roi_info,
        selected_components=metadata["components"],
        mask=mask,
        roi_extraction_parameters=metadata["parameters"] if "parameters" in metadata else None,
        history=history,
        spacing=([10 ** -9] + list(spacing)) if spacing is not None else None,
    )


def empty_fun(_a0=None, _a1=None):
    """
    This is empty fun to pass as callback to for report.
    """


class LoadROI(LoadBase):
    """
    Load ROI segmentation data.
    """

    @classmethod
    def get_name(cls):
        return "Segmentation to image (*.seg *.tgz)"

    @classmethod
    def get_short_name(cls):
        return "seg"

    @staticmethod
    def fix_parameters(profile: ROIExtractionProfile):
        if profile is None:
            return
        if (profile.algorithm in {"Threshold", "Auto Threshold"}) and isinstance(profile.values["smooth_border"], bool):
            if profile.values["smooth_border"] and "smooth_border_radius" in profile.values:
                profile.values["smooth_border"] = {
                    "name": "Opening",
                    "values": {"smooth_border_radius": profile.values["smooth_border_radius"]},
                }
                del profile.values["smooth_border_radius"]
            else:
                profile.values["smooth_border"] = {"name": "None", "values": {}}
        return profile

    @classmethod
    def load(
        cls,
        load_locations: typing.List[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Callable[[int, int], typing.Any] = None,
        step_changed: typing.Callable[[int], typing.Any] = None,
        metadata: typing.Optional[dict] = None,
    ) -> MaskProjectTuple:
        segmentation_tuple = load_stack_segmentation(
            load_locations[0], range_changed=range_changed, step_changed=step_changed
        )
        if segmentation_tuple.roi_extraction_parameters is None:
            parameters = defaultdict(lambda: None)
        else:
            parameters = defaultdict(
                lambda: None,
                [(int(k), cls.fix_parameters(v)) for k, v in segmentation_tuple.roi_extraction_parameters.items()],
            )
        return dataclasses.replace(segmentation_tuple, roi_extraction_parameters=parameters)

    @classmethod
    def partial(cls):
        return False


class LoadROIParameters(LoadBase):
    """
    Load parameters of ROI segmentation. From segmentation file or from json
    """

    @classmethod
    def get_name(cls):
        return "Segmentation parameters (*.json *.seg *.tgz)"

    @classmethod
    def get_short_name(cls):
        return "seg_par"

    @classmethod
    def load(
        cls,
        load_locations: typing.List[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Callable[[int, int], typing.Any] = None,
        step_changed: typing.Callable[[int], typing.Any] = None,
        metadata: typing.Optional[dict] = None,
    ) -> MaskProjectTuple:
        file_data = load_locations[0]

        if isinstance(file_data, (str, Path)):
            ext = os.path.splitext(file_data)[1]
            if ext == ".json":
                project_metadata = load_metadata(file_data)
                if isinstance(metadata, ROIExtractionProfile):
                    parameters = {1: metadata}
                else:
                    parameters = defaultdict(
                        lambda: None,
                        [(int(k), LoadROI.fix_parameters(v)) for k, v in project_metadata["parameters"].items()],
                    )
                return MaskProjectTuple(file_path=file_data, image=None, roi_extraction_parameters=parameters)

        tar_file, _ = open_tar_file(file_data)
        try:
            project_metadata = load_metadata(tar_file.extractfile("metadata.json").read().decode("utf8"))
            parameters = defaultdict(
                lambda: None,
                [(int(k), LoadROI.fix_parameters(v)) for k, v in project_metadata["parameters"].items()],
            )
        finally:
            if isinstance(file_data, (str, Path)):
                tar_file.close()
        return MaskProjectTuple(file_path=file_data, image=None, roi_extraction_parameters=parameters)


class LoadROIImage(LoadBase):
    """
    Load ROI segmentation and image which is pointed in.
    """

    @classmethod
    def get_name(cls):
        return "ROI project with image (*.seg *.tgz)"

    @classmethod
    def get_short_name(cls):
        return "roi_image"

    @classmethod
    def load(
        cls,
        load_locations: typing.List[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Callable[[int, int], typing.Any] = None,
        step_changed: typing.Callable[[int], typing.Any] = None,
        metadata: typing.Optional[dict] = None,
    ) -> MaskProjectTuple:
        seg = LoadROI.load(load_locations)
        base_file = load_locations[1] if len(load_locations) > 1 else seg.image
        if base_file is None:
            raise OSError("base file for segmentation not defined")
        if os.path.isabs(base_file):
            file_path = base_file
        elif not isinstance(load_locations[0], str):
            raise OSError(f"Cannot use relative path {base_file} for non path argument")
        else:
            file_path = os.path.join(os.path.dirname(load_locations[0]), base_file)
        if not os.path.exists(file_path):
            raise OSError(f"Base file for segmentation do not exists: {base_file} -> {file_path}")
        if metadata is None:
            metadata = {"default_spacing": (10 ** -6, 10 ** -6, 10 ** -6)}
        image = GenericImageReader.read_image(
            file_path,
            callback_function=partial(proxy_callback, range_changed, step_changed),
            default_spacing=metadata["default_spacing"],
        )
        # noinspection PyProtectedMember
        # image.file_path = load_locations[0]
        return dataclasses.replace(
            seg, file_path=image.file_path, image=image, roi_info=seg.roi_info.fit_to_image(image)
        )


class LoadStackImage(LoadBase):
    """
    Load image from standard microscopy images
    """

    @classmethod
    def get_name(cls):
        return "Image(*.tif *.tiff *.lsm *.czi *.oib *.oif *.obsep)"

    @classmethod
    def get_short_name(cls):
        return "img"

    @classmethod
    def load(
        cls,
        load_locations: typing.List[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Callable[[int, int], typing.Any] = None,
        step_changed: typing.Callable[[int], typing.Any] = None,
        metadata: typing.Optional[dict] = None,
    ) -> MaskProjectTuple:
        if metadata is None:
            metadata = {"default_spacing": (10 ** -6, 10 ** -6, 10 ** -6)}
        image = GenericImageReader.read_image(
            load_locations[0],
            callback_function=partial(proxy_callback, range_changed, step_changed),
            default_spacing=metadata["default_spacing"],
        )

        return MaskProjectTuple(image.file_path, image, selected_components=[])


class LoadStackImageWithMask(LoadBase):
    """
    Load image, hne mask from secondary file
    """

    @classmethod
    def get_short_name(cls):
        return "img_with_mask"

    @classmethod
    def get_next_file(cls, file_paths: typing.List[str]):
        base, ext = os.path.splitext(file_paths[0])
        return base + "_mask" + ext

    @classmethod
    def number_of_files(cls):
        return 2

    @classmethod
    def load(
        cls,
        load_locations: typing.List[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Callable[[int, int], typing.Any] = None,
        step_changed: typing.Callable[[int], typing.Any] = None,
        metadata: typing.Optional[dict] = None,
    ) -> typing.Union[ProjectInfoBase, typing.List[ProjectInfoBase]]:
        if metadata is None:
            metadata = {"default_spacing": (10 ** -6, 10 ** -6, 10 ** -6)}
        image = GenericImageReader.read_image(
            load_locations[0],
            load_locations[1],
            callback_function=partial(proxy_callback, range_changed, step_changed),
            default_spacing=metadata["default_spacing"],
        )

        return MaskProjectTuple(image.file_path, image, mask=image.mask, selected_components=[])

    @classmethod
    def get_name(cls) -> str:
        return "Image with mask(*.tif *.tiff *.lsm *.czi *.oib *.oif)"


class SaveROI(SaveBase):
    """
    Save current ROI
    """

    @classmethod
    def get_name(cls):
        return "ROI project (*.seg *.tgz)"

    @classmethod
    def get_short_name(cls):
        return "seg"

    @classmethod
    def get_fields(cls):
        return [AlgorithmProperty("relative_path", "Relative Path\nin segmentation", False)]

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info: MaskProjectTuple,
        parameters: dict,
        range_changed=None,
        step_changed=None,
    ):
        save_stack_segmentation(save_location, project_info, parameters)


def save_components(
    image: Image,
    components: list,
    dir_path: str,
    roi_info: ROIInfo,
    parameters=None,
    points: typing.Optional[np.ndarray] = None,
    range_changed=None,
    step_changed=None,
    writer_class: typing.Type[BaseImageWriter] = ImageWriter,
):
    if range_changed is None:
        range_changed = empty_fun
    if step_changed is None:
        step_changed = empty_fun

    if parameters is None:
        parameters = SaveComponents.get_default_values()

    roi_info = roi_info.fit_to_image(image)
    os.makedirs(dir_path, exist_ok=True)

    file_name = os.path.splitext(os.path.basename(image.file_path))[0]
    important_axis = "XY" if image.is_2d else "XYZ"
    index_to_frame_points = image.calc_index_to_frame(image.axis_order, important_axis)
    points_casted = points.astype(np.uint16) if points is not None else None
    if not components:
        components = list(roi_info.bound_info.keys())
    range_changed(0, 2 * len(components))
    for i in components:
        components_mark = np.array(roi_info.roi == i)
        im = image.cut_image(
            components_mark, replace_mask=True, frame=parameters["frame"], zero_out_cut_area=parameters["mask_data"]
        )
        if points is not None and points_casted is not None:
            points_mask = components_mark[tuple(points_casted.T)]
            filtered_points = points[points_mask]
            filtered_points[:, 1] = np.round(filtered_points[:, 1])
            lower_bound = np.min(np.nonzero(components_mark), axis=1)
            for j in index_to_frame_points:
                lower_bound[j] -= FRAME_THICKNESS
            napari_write_points(
                os.path.join(dir_path, f"{file_name}_component{i}.csv"), filtered_points - lower_bound, {}
            )

        # print(f"[run] {im}")
        writer_class.save(im, os.path.join(dir_path, f"{file_name}_component{i}.tif"))
        step_changed(2 * i + 1)
        writer_class.save_mask(im, os.path.join(dir_path, f"{file_name}_component{i}_mask.tif"))
        step_changed(2 * i + 2)


class SaveComponents(SaveBase):
    """
    Save selected components in separated files.
    """

    @classmethod
    def get_short_name(cls):
        return "comp"

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info: MaskProjectTuple,
        parameters: dict,
        range_changed=None,
        step_changed=None,
    ):
        save_components(
            project_info.image,
            project_info.selected_components,
            save_location,
            project_info.roi_info,
            parameters,
            project_info.points,
            range_changed,
            step_changed,
        )

    @classmethod
    def get_name(cls) -> str:
        return "Components"

    @classmethod
    def get_fields(cls) -> typing.List[typing.Union[AlgorithmProperty, str]]:
        return [AlgorithmProperty("frame", "Frame", 2), AlgorithmProperty("mask_data", "Mask data", True)]


class SaveComponentsImagej(SaveBase):
    @classmethod
    def get_short_name(cls):
        return "comp_imagej"

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info: MaskProjectTuple,
        parameters: dict,
        range_changed=None,
        step_changed=None,
    ):
        save_components(
            project_info.image,
            project_info.selected_components,
            save_location,
            project_info.roi_info,
            parameters,
            project_info.points,
            range_changed,
            step_changed,
            writer_class=IMAGEJImageWriter,
        )

    @classmethod
    def get_name(cls) -> str:
        return "Components Imagej Tiff"

    @classmethod
    def get_fields(cls) -> typing.List[typing.Union[AlgorithmProperty, str]]:
        return [AlgorithmProperty("frame", "Frame", 2), AlgorithmProperty("mask_data", "Mask data", True)]


class SaveParametersJSON(SaveBase):
    """
    Save parameters of roi mask segmentation
    """

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info,
        parameters: dict = None,
        range_changed=None,
        step_changed=None,
    ):
        """
        :param save_location: path to save
        :param project_info: data to save in json file
        :param parameters: Not used, keep for satisfy interface
        :param range_changed: Not used, keep for satisfy interface
        :param step_changed: Not used, keep for satisfy interface
        :return:
        """
        with open(save_location, "w", encoding="utf-8") as ff:
            json.dump({"parameters": project_info.roi_extraction_parameters}, ff, cls=ProfileEncoder)

    @classmethod
    def get_fields(cls) -> typing.List[typing.Union[AlgorithmProperty, str]]:
        return []

    @classmethod
    def get_short_name(cls):
        return "json"

    @classmethod
    def get_name(cls) -> str:
        return "Parameters (*.json)"


class LoadROIFromTIFF(LoadBase):
    @classmethod
    def load(
        cls,
        load_locations: typing.List[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Callable[[int, int], typing.Any] = None,
        step_changed: typing.Callable[[int], typing.Any] = None,
        metadata: typing.Optional[dict] = None,
    ) -> typing.Union[ProjectInfoBase, typing.List[ProjectInfoBase]]:
        image = TiffImageReader.read_image(load_locations[0])
        roi = image.get_channel(0)
        return MaskProjectTuple(
            file_path=load_locations[0],
            image=None,
            roi_info=ROIInfo(roi),
            roi_extraction_parameters=defaultdict(lambda: None),
        )

    @classmethod
    def get_short_name(cls):
        return "roi_tiff"

    @classmethod
    def get_name(cls) -> str:
        return "ROI from tiff (*.tif *.tiff)"


def load_metadata(data: typing.Union[str, Path, typing.TextIO]):
    """
    Load metadata saved in json format for segmentation mask
    :param data: path to json file, string with json, or opened file
    :return: restored structures
    """
    return UpdateLoadedMetadataMask.load_json_data(data)


class UpdateLoadedMetadataMask(UpdateLoadedMetadataBase):
    @classmethod
    def update_segmentation_profile(cls, profile_data: ROIExtractionProfile) -> ROIExtractionProfile:
        profile_data = super().update_segmentation_profile(profile_data)
        if profile_data.algorithm in {"Threshold", "Auto Threshold"}:
            if isinstance(profile_data.values["smooth_border"], bool):
                if profile_data.values["smooth_border"]:
                    profile_data.values["smooth_border"] = {
                        "name": "Opening",
                        "values": {"smooth_border_radius": profile_data.values["smooth_border_radius"]},
                    }
                else:
                    profile_data.values["smooth_border"] = {"name": "None", "values": {}}
                if "smooth_border_radius" in profile_data.values:
                    del profile_data.values["smooth_border_radius"]
            if "noise_removal" in profile_data.values:
                profile_data.values["noise_filtering"] = profile_data.values["noise_removal"]
                del profile_data.values["noise_removal"]
        return profile_data


load_dict = Register(
    LoadStackImage, LoadROIImage, LoadStackImageWithMask, LoadPoints, class_methods=LoadBase.need_functions
)
save_parameters_dict = Register(SaveParametersJSON, class_methods=SaveBase.need_functions)
save_components_dict = Register(SaveComponents, SaveComponentsImagej, class_methods=SaveBase.need_functions)
save_segmentation_dict = Register(
    SaveROI, SaveMaskAsTiff, SaveROIAsTIFF, SaveROIAsNumpy, class_methods=SaveBase.need_functions
)



"""
This module contains simple, useful widgets which implementation is to short to create separated files for them
"""


import math
import typing
import warnings
from enum import Enum
from sys import platform

from qtpy import PYQT5
from qtpy.QtCore import QPointF, QRect, Qt, QTimer, Signal
from qtpy.QtGui import QColor, QFontMetrics, QPainter, QPaintEvent, QPalette, QPolygonF
from qtpy.QtWidgets import (
    QAbstractSpinBox,
    QComboBox,
    QDoubleSpinBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QSlider,
    QSpinBox,
    QTextEdit,
    QWidget,
)
from superqt import QEnumComboBox

from PartSegCore.universal_const import UNIT_SCALE, Units

enum_type = Enum if PYQT5 else object


class ChannelComboBox(QComboBox):
    """Combobox for selecting channel index. Channel numeration starts from 1 for user and from 0 for developer"""

    def get_value(self) -> int:
        """Return current channel. Starting from 0"""
        return self.currentIndex()

    def set_value(self, val: int):
        """Set current channel . Starting from 0"""
        self.setCurrentIndex(int(val))

    def change_channels_num(self, num: int):
        """Change number of channels"""
        block = self.blockSignals(True)
        index = self.currentIndex()
        self.clear()
        self.addItems(list(map(str, range(1, num + 1))))
        if index < 0 or index > num:
            index = 0
        self.blockSignals(block)
        self.setCurrentIndex(index)


EnumType = typing.TypeVar("EnumType", bound=Enum)


class EnumComboBox(QComboBox):
    """
    Combobox for choose :py:class:`enum.Enum` values

    :param enum: Enum on which base combo box should be created.
        For proper showing labels overload the ``__str__`` function of given :py:class:`enum.Enum`
    """

    current_choose = Signal(enum_type)
    """:py:class:`Signal` emitted when currentIndexChanged is emitted.
    Argument is selected value"""

    def __init__(self, enum: type(EnumType), parent=None):
        warnings.warn(
            "EnumComboBox is deprecated, use superqt.QEnumComboBox instead", category=DeprecationWarning, stacklevel=2
        )
        super().__init__(parent=parent)
        self.enum = enum
        self.addItems(list(map(str, enum.__members__.values())))
        self.currentIndexChanged.connect(self._emit_signal)

    def get_value(self) -> EnumType:
        """current value as Enum member"""
        return list(self.enum.__members__.values())[self.currentIndex()]

    def _emit_signal(self):
        self.current_choose.emit(self.get_value())

    def set_value(self, value: typing.Union[EnumType, int]):
        """Set value with Eunum or int"""
        if not isinstance(value, (Enum, int)):
            return
        if isinstance(value, Enum):
            self.setCurrentText(str(value))
        else:
            self.setCurrentIndex(value)


class Spacing(QWidget):
    """
    :type elements: list[QDoubleSpinBox | QSpinBox]
    """

    def __init__(
        self,
        title,
        data_sequence,
        unit: Units,
        parent=None,
        input_type=QDoubleSpinBox,
        decimals=2,
        data_range=(0, 100000),
        single_step=1,
    ):
        """
        :type data_sequence: list[(float)]
        :param data_sequence:
        :type input_type: () -> (QDoubleSpinBox | QSpinBox)
        :param parent:
        :type decimals: int|None
        :type data_range: (float, float)
        :type single_step: float
        :type title: str
        """
        super().__init__(parent)
        layout = QHBoxLayout()
        layout.addWidget(QLabel(f"<strong>{title}</strong>"))
        self.elements = []
        if len(data_sequence) == 2:
            data_sequence = (1,) + tuple(data_sequence)
        for name, value in zip(["z", "y", "x"], data_sequence):
            lab = QLabel(name + ":")
            layout.addWidget(lab)
            val = input_type()
            val.setButtonSymbols(QAbstractSpinBox.NoButtons)
            if isinstance(val, QDoubleSpinBox):
                val.setDecimals(decimals)
            val.setRange(*data_range)
            val.setValue(value * UNIT_SCALE[unit.value])
            val.setAlignment(Qt.AlignRight)
            val.setSingleStep(single_step)
            font = val.font()
            fm = QFontMetrics(font)
            val_len = max(fm.width(str(data_range[0])), fm.width(str(data_range[1]))) + fm.width(" " * 8)
            val.setFixedWidth(val_len)
            layout.addWidget(val)
            self.elements.append(val)
        self.units = QEnumComboBox(enum_class=Units)
        self.units.setCurrentEnum(unit)
        layout.addWidget(self.units)
        self.has_units = True
        layout.addStretch(1)
        self.setLayout(layout)

    def get_values(self):
        return [x.value() / UNIT_SCALE[self.units.currentEnum().value] for x in self.elements]

    def set_values(self, value_list):
        for val, wid in zip(value_list, self.elements):
            wid.setValue(val)

    def get_unit_str(self):
        if self.has_units:
            return self.units.currentText()
        return ""


def right_label(text):
    label = QLabel(text)
    label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)
    return label


def set_position(elem, previous, dist=10):
    pos_y = previous.pos().y()
    if platform.system() == "Darwin" and isinstance(elem, QLineEdit):
        pos_y += 3
    if platform.system() == "Darwin" and isinstance(previous, QLineEdit):
        pos_y -= 3
    if platform.system() == "Darwin" and isinstance(previous, QSlider):
        pos_y -= 10
    if platform.system() == "Darwin" and isinstance(elem, QSpinBox):
        pos_y += 7
    if platform.system() == "Darwin" and isinstance(previous, QSpinBox):
        pos_y -= 7
    elem.move(previous.pos().x() + previous.size().width() + dist, pos_y)


def _verify_bounds(bounds):
    if bounds is None:
        return
    try:
        if len(bounds) != 2:
            raise ValueError(f"Wrong bounds format for bounds: {bounds}")
        if not (isinstance(bounds[1], (int, float)) and isinstance(bounds[0], typing.Iterable)):
            raise ValueError(f"Wrong bounds format for bounds: {bounds}")
        for el in bounds[1]:
            if len(el) != 2 or not isinstance(el[0], (int, float)) or not isinstance(el[1], (int, float)):
                raise ValueError(f"Wrong bounds format for bounds: {bounds}")
    except TypeError:
        raise ValueError(f"Wrong bounds format for bounds: {bounds}")


class CustomSpinBox(QSpinBox):
    """
    Spin box for integer with dynamic single steep

    :param bounds: Bounds for changing single step. Default value:
        ``((300, 1), (1000, 10), (10000, 100)), 1000``
        Format:
        ``(List[(threshold, single_step)], default_single_step)``
        the single_step is chosen by checking upper bound of threshold of
        current spin box value
    """

    def __init__(self, *args, bounds=None, **kwargs):
        _verify_bounds(bounds)
        super().__init__(*args, **kwargs)
        self.valueChanged.connect(self._value_changed)
        if bounds is None:
            self.bounds = ((300, 1), (1000, 10), (10000, 100)), 1000
        else:
            self.bounds = bounds

    def _value_changed(self, val: int):
        val = abs(val)
        for el in self.bounds[0]:
            if val < el[0]:
                self.setSingleStep(el[1])
                break
        else:
            self.setSingleStep(self.bounds[1])


class CustomDoubleSpinBox(QDoubleSpinBox):
    """
    Spin box for float with dynamic single steep

    :param bounds: Bounds for changing single step. Default value:
        ``((300, 1), (1000, 10), (10000, 100)), 1000``
        Format:
        ``(List[(threshold, single_step)], default_single_step)``
        the single_step is chosen by checking upper bound of threshold of
        current spin box value
    """

    def __init__(self, *args, bounds=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.valueChanged.connect(self._value_changed)
        if bounds is None:
            self.bounds = ((0.2, 0.01), (2, 0.1), (300, 1), (1000, 10), (10000, 100)), 1000
        else:
            self.bounds = bounds

    def _value_changed(self, val: float):
        val = abs(val)
        for el in self.bounds[0]:
            if val < el[0]:
                self.setSingleStep(el[1])
                break
        else:
            self.setSingleStep(self.bounds[1])


class ProgressCircle(QWidget):
    """
    This is widget for generating circuital progress bar

    :param background: color of background circle, need to be acceptable by :py:class:`PyQt5.QtGui.QColor` constructor.
    :param main_color: color of progress marking, need to be acceptable by :py:class:`PyQt5.QtGui.QColor` constructor.

    .. warning::
      This widget currently have no minimum size. You need to specify it in your code
    """

    def __init__(self, background: QColor = "white", main_color: QColor = "darkCyan", parent=None):
        super().__init__(parent)
        self.nominator = 1
        self.denominator = 1
        self.background_color = QColor(background)
        self.main_color = QColor(main_color)

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)
        painter.save()
        painter.setRenderHint(QPainter.Antialiasing)
        size = min(self.width(), self.height())
        rect = QRect(0, 0, size, size)
        painter.setBrush(self.background_color)
        painter.setPen(self.background_color)
        painter.drawEllipse(rect)
        painter.setBrush(self.main_color)
        painter.setPen(self.main_color)
        factor = self.nominator / self.denominator
        radius = size / 2
        if factor > 0.5:
            painter.drawChord(rect, 0, 16 * 360 * 0.5)
            painter.drawChord(rect, 16 * 180, 16 * 360 * (factor - 0.5))
            zero_point = QPointF(0, radius)
        else:
            painter.drawChord(rect, 0, 16 * 360 * factor)
            zero_point = QPointF(size, radius)
        mid_point = QPointF(radius, radius)
        point = mid_point + QPointF(
            math.cos(math.pi * (factor * 2)) * radius, -math.sin(math.pi * (factor * 2)) * radius
        )
        polygon = QPolygonF()
        polygon += mid_point
        polygon += zero_point
        polygon += point
        painter.drawPolygon(polygon)
        painter.restore()

    def set_fraction(self, nominator, denominator=1):
        """
        Set fraction.

        :param nominator: of fraction if denominator is set to 1 then nominator should be from range [0, 1]
        :param denominator: as name
        """
        self.nominator = nominator
        self.denominator = denominator
        self.repaint()


class InfoLabel(QWidget):
    """
    Label for cyclic showing text from list. It uses :py:class:`~.ProgressCircle`
    to inform user abu time to change text.

    :param text_list: texts to be in cyclic use
    :param delay: time in milliseconds between changes
    :param parent: passed to :py:class:`QWidget` constructor
    """

    def __init__(self, text_list: typing.List[str], delay: int = 10000, parent=None):
        if len(text_list) == 0:
            raise ValueError("List of text to show should be non empty.")
        super().__init__(parent)
        self.text_list = text_list
        self.index = 0
        self.delay = delay
        self.time = 0
        self.time_step = 100
        self.timer = QTimer()
        self.timer.setSingleShot(False)
        self.timer.timeout.connect(self.one_step)
        self.timer.setInterval(self.time_step)
        self.setToolTip("Double click for change suggestion. Hold mouse over for stop changes.")
        self.label = QLabel()
        self.progress = ProgressCircle()
        self.progress.setFixedHeight(25)
        self.progress.setFixedWidth(25)
        layout = QHBoxLayout()
        layout.addWidget(self.progress)
        layout.addWidget(self.label)
        self.setLayout(layout)
        self.change_text()

    def mouseDoubleClickEvent(self, _):
        """Force change text"""
        self.change_text()
        self.time = 0
        self.progress.set_fraction(self.time, self.delay)

    def enterEvent(self, _):
        """stop timer"""
        self.timer.stop()

    def leaveEvent(self, _):
        """Start timer"""
        self.timer.start()

    def showEvent(self, _):
        """Start timer"""
        self.timer.start()

    def hideEvent(self, _):
        """Stop timer"""
        self.timer.stop()

    def one_step(self):
        self.time += self.time_step
        if self.time > self.delay:
            self.time = 0
            self.change_text()
        self.progress.set_fraction(self.time, self.delay)

    def change_text(self):
        """Change text in cyclic mode"""
        self.label.setText(self.text_list[self.index])
        self.index = (self.index + 1) % len(self.text_list)


class TextShow(QTextEdit):
    """
    Show text with word wrap and scroll if needed.
    Limit to show first :py:attr:`lines` without scroll bar.
    """

    def __init__(self, lines=5, text="", parent=None):
        super().__init__(text, parent)
        self.lines = lines
        self.setReadOnly(True)
        # self.setTextBackgroundColor()
        p: QPalette = self.palette()
        p.setColor(QPalette.Base, p.color(self.backgroundRole()))
        self.setPalette(p)

    def height(self):
        metrics = QFontMetrics(self.currentFont())
        height = metrics.height()
        return height * 5

    def sizeHint(self):
        s = super().sizeHint()
        metrics = QFontMetrics(self.currentFont())
        height = metrics.height()
        s.setHeight(int(height * (self.lines + 0.5)))
        return s



changelog = """
# Changelog

## 0.13.15

### Bug Fixes

-   Using `translation` instead of `translation_grid` for shifting layers. (#474)
-   Bugs in napari plugins (#478)
-   Missing mask when using roi extraction from napari (#479)
-   Fix segmentation fault on macos machines (#487)
-   Fixes for napari 0.4.13 (#506)

### Documentation

-   Create 0.13.15 release (#511)
-   Add categories and preview page workflow for the napari hub (#489)

### Features

-   Assign properties to mask layer in napari measurement widget (#480)

### Build

-   Bump qtpy from 1.11.3 to 2.0.0 in /requirements (#498)
-   Bump pydantic from 1.8.2 to 1.9.0 in /requirements (#496)
-   Bump sentry-sdk from 1.5.1 to 1.5.2 in /requirements (#497)
-   Bump sphinx from 4.3.1 to 4.3.2 in /requirements (#500)
-   Bump pyinstaller from 4.7 to 4.8 in /requirements (#502)
-   Bump pillow from 8.4.0 to 9.0.0 in /requirements (#501)
-   Bump requests from 2.26.0 to 2.27.1 in /requirements (#495)
-   Bump numpy from 1.21.4 to 1.22.0 in /requirements (#499)
-   Bump numpy from 1.22.0 to 1.22.1 in /requirements (#509)
-   Bump sphinx from 4.3.2 to 4.4.0 in /requirements (#510)

## 0.13.14

### Bug Fixes

-   ROI alternative representation (#471)
-   Change additive to translucent in rendering ROI and Mask (#472)

### Features

-   Add morphological watershed segmentation (#469)
-   Add Bilateral image filter (#470)

## 0.13.13

### Bug Fixes

-   Fix bugs in the generation process of the changelog for release. (#428)
-   Restoring ROI on home button click in compare viewer (#443)
-   Fix Measurement name prefix in bundled PartSeg. (#458)
-   Napari widgets registration in pyinstaller bundle (#465)
-   Hide points button if no points are loaded, hide Mask checkbox if no mask is set (#463)
-   Replace Label data instead of adding/removing layers - fix blending layers (#464)

### Features

-   Add threshold information in layer annotation in the Multiple Otsu ROI extraction method (#430)
-   Add option to select rendering method for ROI (#431)
-   Add callback mechanism to ProfileDict, live update of ROI render parameters (#432)
-   Move the info bar on the bottom of the viewer (#442)
-   Add options to load recent files in multiple files widget (#444)
-   Add ROI annotations as properties to napari labels layer created by ROI Extraction widgets (#445)
-   Add signals to ProfileDict, remove redundant synchronization mechanisms (#449)
-   Allow ignoring updates for 21 days (#453)
-   Save all components if no components selected in mask segmentation (#456)
-   Add modal dialog for search ROI components (#459)
-   Add full measurement support as napari widget (#460)
-   Add search labels as napari widget (#467)

### Refactor

-   Export common code for load/save dialog to one place (#437)
-   Change most of call QFileDialog to more generic code (#440)

### Testing

-   Add test for `PartSeg.common_backend` module (#433)

## 0.13.12

### Bug Fixes

-   Importing the previous version of settings (#406)
-   Cutting without masking data (#407)
-   Save in subdirectory in batch plan (#414)
-   Loading plugins for batch processing (#423)

### Features

-   Add randomization option for correlation calculation (#421)
-   Add Imagej TIFF writter for image. (#405)
-   Mask create widget for napari (#395)
-   In napari roi extraction method show information from roi extraction method (#408)
-   Add `*[0-9].tif` button in batch processing window (#412)
-   Better label representation in 3d view (#418)

### Refactor

-   Use Font Awesome instead of custom symbols (#424)

## 0.13.11

### Bug Fixes

-   Adding mask in Prepare Plan for batch (#383)
-   Set proper completion mode in SearchComboBox (#384)
-   Showing warnings on the error with ROI load (#385)

### Features

-   Add CellFromNucleusFlow "Cell from nucleus flow" cell segmentation method  (#367)
-   When cutting components in PartSeg ROI mask allow not masking outer data (#379)
-   Theme selection in GUI (#381)
-   Allow return points from ROI extraction algorithm (#382)
-   Add measurement to get ROI annotation by name. (#386)
-   PartSeg ROI extraction algorithms as napari plugins (#387)
-   Add  Pearson, Mander's, Intensity, Spearman colocalization measurements (#392)
-   Separate standalone napari settings from PartSeg embedded napari settings (#397)

### Performance

-   Use faster calc bound function (#375)

### Refactor

-   Remove CustomApplication (#389)

## 0.13.10

-   change tiff save backend to ome-tiff
-   add `DistanceROIROI` and `ROINeighbourhoodROI` measurements

## 0.13.9

-   annotation show bugfix

## 0.13.8

-   napari deprecation fixes
-   speedup simple measurement
-   bundle plugins initial support

## 0.13.7

-   add measurements widget for napari
-   fix bug in pipeline usage

## 0.13.6

-   Hotfix release
-   Prepare for a new napari version

## 0.13.5

-   Small fixes for error reporting
-   Fix mask segmentation

## 0.13.4

-   Bugfix for outdated profile/pipeline preview

## 0.13.3

-   Fix saving roi_info in multiple files and history

## 0.13.2

-   Fix showing label in select label tab

## 0.13.1

-   Add Haralick measurements
-   Add obsep file support

## 0.13.0

-   Add possibility of custom input widgets for algorithms
-   Switch to napari Colormaps instead of custom one
-   Add points visualization
-   Synchronization widget for builtin (View menu) napari viewer
-   Drop Python 3.6

## 0.12.7

-   Fixes for napari 0.4.6

## 0.12.6

-   Fix prev_mask_get
-   Fix cache mechanism on mask change
-   Update PyInstaller build

## 0.12.5

-   Fix bug in pipeline execute

## 0.12.4

-   Fix ROI Mask windows related build (signal not properly connected)

## 0.12.3

-   Fix ROI Mask

## 0.12.2

-   Fix windows bundle

## 0.12.1

-   History of last opened files
-   Add ROI annotation and ROI alternatives
-   Minor bugfix

## 0.12.0

-   Toggle multiple files widget in View menu
-   Toggle Left panel in ROI Analysis in View Menu
-   Rename Mask Segmentation to ROI Mask
-   Add documentation for interface
-   Add Batch processing tutorial
-   Add information about errors to batch processing output file
-   Load image from the batch prepare window
-   Add search option in part of list and combo boxes
-   Add drag and drop mechanism to load list of files to batch window.

## 0.11.5

-   add side view to viewer
-   fix horizontal view for Measurements result table

## 0.11.4

-   bump to napari 0.3.8 in bundle
-   fix bug with not presented segmentation loaded from project
-   add frame (1 pix) to image cat from base one based on segmentation
-   pin to Qt version to 5.14

## 0.11.3

-   prepare for napari 0.3.7
-   split napari io plugin on multiple part
-   better reporting for numpy array via sentry
-   fix setting color for mask marking

## 0.11.2

-   Speedup image set in viewer using async calls
-   Fix bug in long name of sheet with parameters

## 0.11.1

-   Add screenshot option in View menu
-   Add Voxels measurements

## 0.11.0

-   Make sprawl algorithm name shorter
-   Unify capitalisation of measurement names
-   Add simple measurements to mask segmentation
-   Use napari as viewer
-   Add possibility to preview additional output of algorithms (In View menu)
-   Update names of available Algorithm and Measurement to be more descriptive.

## 0.10.8

-   fix synchronisation between viewers in Segmentation Analysis
-   fix batch crash on error during batch run, add information about file on which calculation fails
-   add changelog preview in Help > About

## 0.10.7

-   in measurements, on empty list of components mean will return 0

## 0.10.6

-   fix border rim preview
-   fix problem with size of image preview
-   zoom with scroll and moving if rectangle zoom is not marked

## 0.10.5

-   make PartSeg PEP517 compatible.
-   fix multiple files widget on Windows (path normalisation)

## 0.10.4

-   fix slow zoom

## 0.10.3

-   deterministic order of elements in batch processing.

## 0.10.2

-   bugfixes

## 0.10.1

-   bugfixes

## 0.10.0

-   Add creating custom label coloring.
-   Change execs interpreter to python 3.7.
-   Add masking operation in Segmentation Mask.
-   Change license to BSD.
-   Allow select root type in batch processing.
-   Add median filter in preview.

## 0.9.7

-   fix bug in compare mask

## 0.9.6

-   fix bug in loading project with mask
-   upgrade PyInstaller version (bug  GHSA-7fcj-pq9j-wh2r)

## 0.9.5

-   fix bug in loading project in "Segmentation analysis"

## 0.9.4

-   read mask segmentation projects
-   choose source type in batch
-   add initial support to OIF and CZI file format
-   extract utils to PartSegCore module
-   add automated tests of example notebook
-   reversed mask
-   load segmentation parameters in mask segmentation
-   allow use sprawl in segmentation tool
-   add radial split of mask for measurement
-   add all measurement results in batch, per component sheet

## 0.9.3

-   start automated build documentation
-   change color map backend and allow for user to create custom color map.
-   segmentation compare
-   update test engines
-   support of PySide2

## 0.9.2.3

-   refactor code to make easier create plugin for mask segmentation
-   create class base updater for update outdated algorithm description
-   fix save functions
-   fix different bugs

## 0.9.2.2

-   extract static data to separated package
-   update marker of fix range and add mark of gauss in channel control

## 0.9.2.1

-   add VoteSmooth and add choosing of smooth algorithm

## 0.9.2

-   add pypi base check for update

-   remove resetting image state when change state in same image

-   in stack segmentation add options to picking components from segmentation's

-   in mask segmentation add:

    -   preview of segmentation parameters per component,
    -   save segmentation parameters in save file
    -   new implementation of batch mode.

## 0.9.1

-   Add multiple files widget

-   Add Calculating distances between segmented object and mask

-   Batch processing plan fixes:

    -   Fix adding pipelines to plan
    -   Redesign mask widget

-   modify measurement backend to allow calculate multi channel measurements.

## 0.9

Begin of changelog
"""



import collections
import inspect
import typing
from contextlib import suppress
from copy import deepcopy
from enum import Enum

import magicgui
import numpy as np
from magicgui.widgets import ComboBox, Widget, create_widget
from napari.layers.base import Layer
from packaging.version import parse as parse_version
from qtpy.QtCore import Signal
from qtpy.QtGui import QHideEvent, QPainter, QPaintEvent, QResizeEvent
from qtpy.QtWidgets import (
    QAbstractSpinBox,
    QApplication,
    QCheckBox,
    QComboBox,
    QFormLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QScrollArea,
    QStackedLayout,
    QVBoxLayout,
    QWidget,
)
from superqt import QEnumComboBox

from PartSeg.common_gui.error_report import ErrorDialog
from PartSegCore.algorithm_describe_base import AlgorithmDescribeBase, AlgorithmProperty, ROIExtractionProfile
from PartSegCore.channel_class import Channel
from PartSegCore.segmentation.algorithm_base import (
    ROIExtractionAlgorithm,
    ROIExtractionResult,
    SegmentationLimitException,
)
from PartSegImage import Image

from ..common_backend.base_settings import BaseSettings
from ..common_backend.segmentation_thread import SegmentationThread
from .universal_gui_part import ChannelComboBox, CustomDoubleSpinBox, CustomSpinBox


def update(d, u):
    if not isinstance(d, typing.MutableMapping):
        d = {}
    for k, v in u.items():
        d[k] = update(d.get(k, {}), v) if isinstance(v, collections.abc.Mapping) else v
    return d


def _pretty_print(data, indent=2) -> str:
    if isinstance(data, typing.Mapping):
        res = "\n"
        for k, v in data.items():
            res += f"{' ' * indent}{k}: {_pretty_print(v, indent+2)}\n"
        return res[:-1]
    return str(data)


class ProfileSelect(QComboBox):
    def __init__(self):
        super().__init__()
        self._settings = None

    def add_settings(self, settings: BaseSettings):
        self._settings = settings
        if hasattr(settings, "roi_profiles"):
            self.addItems(settings.roi_profiles.keys())

    def get_value(self):
        if self._settings is not None and hasattr(self._settings, "roi_profiles"):
            return self._settings.roi_profiles[self.currentText()]
        return None

    def set_value(self, value):
        if (
            self._settings is not None
            and hasattr(self._settings, "roi_profiles")
            and value.name in self._settings.roi_profiles
        ):
            self.setCurrentText(value.name)


class QtAlgorithmProperty(AlgorithmProperty):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._widget = self._get_field()
        self.change_fun = self.get_change_signal(self._widget)
        self._getter, self._setter = self.get_getter_and_setter_function(self._widget)
        with suppress(TypeError, ValueError):
            self._setter(self._widget, self.default_value)

    def get_value(self):
        return self._getter(self._widget)

    def recursive_get_values(self):
        if isinstance(self._widget, SubAlgorithmWidget):
            return self._widget.recursive_get_values()
        return self.get_value()

    def set_value(self, val):
        """set value of widget"""
        with suppress(TypeError, ValueError):
            return self._setter(self._widget, val)

    def get_field(self) -> QWidget:
        """
        Get representing widget
        :return:
        :rtype:
        """
        return self._widget

    @classmethod
    def from_algorithm_property(cls, ob):
        """
        Create class instance base on :py:class:`.AlgorithmProperty` instance

        :type ob: AlgorithmProperty | str
        :param ob: AlgorithmProperty object or label
        :return: QtAlgorithmProperty | QLabel
        """
        if isinstance(ob, AlgorithmProperty):
            return cls(
                name=ob.name,
                user_name=ob.user_name,
                default_value=ob.default_value,
                options_range=ob.range,
                single_steep=ob.single_step,
                value_type=ob.value_type,
                possible_values=ob.possible_values,
                help_text=ob.help_text,
                per_dimension=ob.per_dimension,
            )
        if isinstance(ob, str):
            return QLabel(ob)
        raise ValueError(f"unknown parameter type {type(ob)} of {ob}")

    @staticmethod
    def _get_numeric_field(ap: AlgorithmProperty):
        if issubclass(ap.value_type, int):
            res = CustomSpinBox()
            if not isinstance(ap.default_value, int):
                raise ValueError(
                    f"Incompatible types. default_value should be type of int. Is {type(ap.default_value)}"
                )
        else:  # issubclass(ap.value_type, float):
            res = CustomDoubleSpinBox()
            if not isinstance(ap.default_value, float):
                raise ValueError(
                    f"Incompatible types. default_value should be type of float. Is {type(ap.default_value)}"
                )
        if ap.default_value < res.minimum():
            res.setMinimum(ap.default_value)
        if ap.range is not None:
            res.setRange(*ap.range)
        return res

    @classmethod
    def _get_field_from_value_type(cls, ap: AlgorithmProperty):
        if issubclass(ap.value_type, Channel):
            res = ChannelComboBox()
            res.change_channels_num(10)
        elif issubclass(ap.value_type, AlgorithmDescribeBase):
            res = SubAlgorithmWidget(ap)
        elif issubclass(ap.value_type, bool):
            res = QCheckBox()
        elif issubclass(ap.value_type, (float, int)):
            res = cls._get_numeric_field(ap)
        elif issubclass(ap.value_type, Enum):
            # noinspection PyTypeChecker
            res = QEnumComboBox(enum_class=ap.value_type)
        elif issubclass(ap.value_type, str):
            res = QLineEdit()
        elif issubclass(ap.value_type, ROIExtractionProfile):
            res = ProfileSelect()
        elif issubclass(ap.value_type, list):
            res = QComboBox()
            res.addItems(list(map(str, ap.possible_values)))
        else:
            res = create_widget(annotation=ap.value_type, options={})
        return res

    def _get_field(self) -> typing.Union[QWidget, Widget]:
        """
        Get proper widget for given field type. Overwrite if would like to support new data types.
        """
        if self.per_dimension:
            self.per_dimension = False
            prop = self.from_algorithm_property(self)
            self.per_dimension = True
            res = ListInput(prop, 3)
        elif not inspect.isclass(self.value_type):
            res = create_widget(annotation=self.value_type, options={})
        elif hasattr(self.value_type, "get_object"):
            res = self.value_type.get_object()
        else:
            res = self._get_field_from_value_type(self)
        tool_tip_text = self.help_text or ""
        tool_tip_text += f" default value: {_pretty_print(self.default_value)}"
        if isinstance(res, QWidget):
            res.setToolTip(tool_tip_text)
        if isinstance(res, Widget):
            res.tooltip = tool_tip_text  # pylint: disable=W0201 # false positive
        return res

    @staticmethod
    def get_change_signal(widget: typing.Union[QWidget, Widget]):
        if isinstance(widget, Widget):
            return widget.changed
        if isinstance(widget, QComboBox):
            return widget.currentIndexChanged
        if isinstance(widget, QCheckBox):
            return widget.stateChanged
        if isinstance(widget, QAbstractSpinBox):
            return widget.valueChanged
        if isinstance(widget, QLineEdit):
            return widget.textChanged
        if isinstance(widget, SubAlgorithmWidget):
            return widget.values_changed
        if isinstance(widget, ListInput):
            return widget.change_signal
        if hasattr(widget, "values_changed"):
            return widget.values_changed
        raise ValueError(f"Unsupported type: {type(widget)}")

    @staticmethod
    def get_getter_and_setter_function(
        widget: typing.Union[QWidget, Widget],
    ) -> typing.Tuple[
        typing.Callable[
            [
                QWidget,
            ],
            typing.Any,
        ],
        typing.Callable[[QWidget, typing.Any], None],  # noqa E231
    ]:
        """
        For each widget type return proper functions. This functions need instance as first argument

        :return: (getter, setter)
        """
        if isinstance(widget, Widget):
            return _value_get, _value_set
        if isinstance(widget, ProfileSelect):
            return widget.__class__.get_value, widget.__class__.set_value
        if isinstance(widget, ChannelComboBox):
            return widget.__class__.get_value, widget.__class__.set_value
        if isinstance(widget, QEnumComboBox):
            return widget.__class__.currentEnum, widget.__class__.setCurrentEnum
        if isinstance(widget, QComboBox):
            return widget.__class__.currentText, widget.__class__.setCurrentText
        if isinstance(widget, QCheckBox):
            return widget.__class__.isChecked, widget.__class__.setChecked
        if isinstance(widget, CustomSpinBox):
            return widget.__class__.value, widget.__class__.setValue
        if isinstance(widget, CustomDoubleSpinBox):
            return widget.__class__.value, widget.__class__.setValue
        if isinstance(widget, QLineEdit):
            return widget.__class__.text, widget.__class__.setText
        if isinstance(widget, SubAlgorithmWidget):
            return widget.__class__.get_values, widget.__class__.set_values
        if isinstance(widget, ListInput):
            return widget.__class__.get_value, widget.__class__.set_value
        if hasattr(widget, "get_value") and hasattr(widget, "set_value"):
            return widget.__class__.get_value, widget.__class__.set_value
        raise ValueError(f"Unsupported type: {type(widget)}")


class ListInput(QWidget):
    change_signal = Signal()

    def __init__(self, property_el: QtAlgorithmProperty, length):
        super().__init__()
        self.input_list = [property_el.from_algorithm_property(property_el) for _ in range(length)]
        layout = QVBoxLayout()
        for el in self.input_list:
            el.change_fun.connect(self.change_signal.emit)
            layout.addWidget(el.get_field())
        self.setLayout(layout)

    def get_value(self):
        return [x.get_value() for x in self.input_list]

    def set_value(self, value):
        if not isinstance(value, (list, tuple)):
            value = [value for _ in range(len(self.input_list))]
        for f, val in zip(self.input_list, value):
            f.set_value(val)


def _any_arguments(fun):
    if parse_version(magicgui.__version__) >= parse_version("0.3.0"):

        def _any():
            fun()

    else:

        def _any(*_):
            fun()

    return _any


class FormWidget(QWidget):
    value_changed = Signal()

    def __init__(
        self,
        fields: typing.List[AlgorithmProperty],
        start_values=None,
        dimension_num=1,
        settings: typing.Optional[BaseSettings] = None,
    ):
        super().__init__()
        if start_values is None:
            start_values = {}
        self.widgets_dict: typing.Dict[str, QtAlgorithmProperty] = {}
        self.channels_chose: typing.List[typing.Union[ChannelComboBox, SubAlgorithmWidget]] = []
        layout = QFormLayout()
        layout.setContentsMargins(10, 0, 10, 0)
        # layout.setVerticalSpacing(0)
        element_list = self._element_list(fields)
        for el in element_list:
            if isinstance(el, QLabel):
                layout.addRow(el)
                continue
            self._add_to_layout(layout, el, start_values, settings)
        self.setLayout(layout)
        self.value_changed.connect(self.update_size)

    def _add_to_layout(self, layout, ap: QtAlgorithmProperty, start_values: typing.MutableMapping, settings):
        label = QLabel(ap.user_name)
        if ap.help_text:
            label.setToolTip(ap.help_text)
        self.widgets_dict[ap.name] = ap
        ap.change_fun.connect(_any_arguments(self.value_changed.emit))
        if isinstance(ap.get_field(), SubAlgorithmWidget):
            layout.addRow(label, ap.get_field().choose)
            layout.addRow(ap.get_field())
            if ap.name in start_values:
                ap.get_field().set_starting(start_values[ap.name])
            ap.change_fun.connect(_any_arguments(self.value_changed.emit))
            return
        if isinstance(ap.get_field(), Widget):
            layout.addRow(label, ap.get_field().native)
            return
        layout.addRow(label, ap.get_field())
        # noinspection PyUnresolvedReferences
        if issubclass(ap.value_type, Channel):
            # noinspection PyTypeChecker
            self.channels_chose.append(ap.get_field())
        if issubclass(ap.value_type, ROIExtractionProfile):
            # noinspection PyTypeChecker
            ap.get_field().add_settings(settings)
        if ap.name in start_values:
            with suppress(KeyError, ValueError, TypeError):
                ap.set_value(start_values[ap.name])

    @staticmethod
    def _element_list(fields) -> typing.Iterable[QtAlgorithmProperty]:
        return map(QtAlgorithmProperty.from_algorithm_property, fields)

    def has_elements(self):
        return len(self.widgets_dict) > 0

    def update_size(self):
        self.setMinimumHeight(self.layout().minimumSize().height())

    def get_values(self):
        return {name: el.get_value() for name, el in self.widgets_dict.items()}

    def recursive_get_values(self):
        return {name: el.recursive_get_values() for name, el in self.widgets_dict.items()}

    def set_values(self, values: dict):
        for name, value in values.items():
            if name in self.widgets_dict:
                self.widgets_dict[name].set_value(value)

    def image_changed(self, image: Image):
        if not image:
            return
        for channel_widget in self.channels_chose:
            if isinstance(channel_widget, ChannelComboBox):
                channel_widget.change_channels_num(image.channels)
            else:
                channel_widget.change_channels_num(image)


class SubAlgorithmWidget(QWidget):
    values_changed = Signal()

    def __init__(self, algorithm_property: AlgorithmProperty):
        super().__init__()
        if not isinstance(algorithm_property.possible_values, typing.MutableMapping):
            raise ValueError(
                "algorithm_property.possible_values should be dict." f"It is {type(algorithm_property.possible_values)}"
            )
        if not isinstance(algorithm_property.default_value, str):
            raise ValueError(
                "algorithm_property.default_value should be str." f"It is {type(algorithm_property.default_value)}"
            )
        self.starting_values = {}
        self.property = algorithm_property
        self.widgets_dict: typing.Dict[str, FormWidget] = {}
        # TODO protect for recursion
        widget = FormWidget(algorithm_property.possible_values[algorithm_property.default_value].get_fields())
        widget.layout().setContentsMargins(0, 0, 0, 0)
        widget.value_changed.connect(self.values_changed)

        self.widgets_dict[algorithm_property.default_value] = widget
        self.choose = QComboBox(self)
        self.choose.addItems(list(algorithm_property.possible_values.keys()))
        self.setContentsMargins(0, 0, 0, 0)

        self.choose.setCurrentText(algorithm_property.default_value)

        self.choose.currentTextChanged.connect(self.algorithm_choose)
        # self.setStyleSheet("border: 1px solid red")
        layout = QVBoxLayout()
        layout.setContentsMargins(4, 4, 4, 4)
        layout.addWidget(widget)
        if not widget.has_elements():
            widget.hide()
            self.hide()
        tmp_widget = QWidget(self)
        # tmp_widget.setMinimumHeight(5000)
        layout.addWidget(tmp_widget)
        self.tmp_widget = tmp_widget
        self.setLayout(layout)

    def set_starting(self, starting_values):
        self.starting_values = starting_values
        # self.set_values(starting_values)

    def set_values(self, val: typing.Mapping):
        if not isinstance(val, typing.Mapping):
            return
        self.choose.setCurrentText(val["name"])
        if val["name"] not in self.widgets_dict:
            self.algorithm_choose(val["name"])
        if val["name"] in self.widgets_dict:
            self.widgets_dict[val["name"]].set_values(val["values"])

    def recursive_get_values(self):
        return {name: el.recursive_get_values() for name, el in self.widgets_dict.items()}

    def get_values(self):
        name = self.choose.currentText()
        values = self.widgets_dict[name].get_values()
        return {"name": name, "values": values}

    def change_channels_num(self, image: Image):
        for i in range(self.layout().count()):
            el = self.layout().itemAt(i)
            if el.widget() and isinstance(el.widget(), FormWidget):
                el.widget().image_changed(image)

    def algorithm_choose(self, name):
        if name not in self.widgets_dict:
            if name not in self.property.possible_values:
                return
            start_dict = {} if name not in self.starting_values else self.starting_values[name]
            try:
                self.widgets_dict[name] = FormWidget(
                    self.property.possible_values[name].get_fields(), start_values=start_dict
                )
            except KeyError as e:
                raise e
            self.widgets_dict[name].layout().setContentsMargins(0, 0, 0, 0)
            self.layout().addWidget(self.widgets_dict[name])
            self.widgets_dict[name].value_changed.connect(self.values_changed)
        widget = self.widgets_dict[name]
        for i in range(self.layout().count()):
            lay_elem = self.layout().itemAt(i)
            if lay_elem.widget():
                lay_elem.widget().hide()
        if widget.has_elements():
            self.show()
            widget.show()
        else:
            self.hide()
        self.values_changed.emit()

    def showEvent(self, _event):
        # workaround for changing size
        self.tmp_widget.hide()

    def paintEvent(self, event: QPaintEvent):
        name = self.choose.currentText()
        if self.widgets_dict[name].has_elements() and event.rect().top() == 0 and event.rect().left() == 0:
            painter = QPainter(self)
            painter.drawRect(event.rect())


class BaseAlgorithmSettingsWidget(QScrollArea):
    values_changed = Signal()
    algorithm_thread: SegmentationThread

    def __init__(self, settings: BaseSettings, name, algorithm: typing.Type[ROIExtractionAlgorithm]):
        """
        For algorithm which works on one channel
        """
        super().__init__()
        self.settings = settings
        self.widget_list = []
        self.name = name
        self.algorithm = algorithm
        main_layout = QVBoxLayout()
        self.info_label = QLabel()
        self.info_label.setHidden(True)
        main_layout.addWidget(self.info_label)
        start_values = settings.get(f"algorithm_widget_state.{name}", {})
        self.form_widget = self._form_widget(algorithm, start_values=start_values)
        self.form_widget.value_changed.connect(self.values_changed.emit)
        # self.form_widget.setMinimumHeight(1500)
        self.setWidget(self.form_widget)
        value_dict = self.settings.get(f"algorithms.{self.name}", {})
        self.set_values(value_dict)
        # self.settings.image_changed[Image].connect(self.image_changed)
        self.algorithm_thread = SegmentationThread(algorithm())
        self.algorithm_thread.info_signal.connect(self.show_info)
        self.algorithm_thread.exception_occurred.connect(self.exception_occurred)

    @staticmethod
    def _form_widget(algorithm, start_values) -> FormWidget:
        return FormWidget(algorithm.get_fields(), start_values=start_values)

    @staticmethod
    def exception_occurred(exc: Exception):
        if isinstance(exc, SegmentationLimitException):
            mess = QMessageBox()
            mess.setIcon(QMessageBox.Critical)
            mess.setText("During segmentation process algorithm meet limitations:\n" + "\n".join(exc.args))
            mess.setWindowTitle("Segmentation limitations")
            mess.exec_()
            return
        if isinstance(exc, RuntimeError) and exc.args[0].startswith(
            "Exception thrown in SimpleITK KittlerIllingworthThreshold"
        ):
            mess = QMessageBox()
            mess.setIcon(QMessageBox.Critical)
            mess.setText("Fail to apply Kittler Illingworth to current data\n" + exc.args[0].split("\n")[1])
            mess.setWindowTitle("Segmentation limitations")
            mess.exec_()
            return
        dial = ErrorDialog(exc, "Error during segmentation", f"{QApplication.instance().applicationName()}")
        dial.exec_()

    def show_info(self, text):
        self.info_label.setText(text)
        self.info_label.setVisible(True)

    def image_changed(self, image: Image):
        self.form_widget.image_changed(image)
        self.algorithm_thread.algorithm.set_image(image)

    def set_mask(self, mask):
        self.algorithm_thread.algorithm.set_mask(mask)

    def mask(self) -> typing.Optional[np.ndarray]:
        return self.algorithm_thread.algorithm.mask

    def set_values(self, values_dict):
        self.form_widget.set_values(values_dict)

    def get_values(self):
        return self.form_widget.get_values()

    def channel_num(self):
        return self.channels_chose.currentIndex()

    def execute(self, exclude_mask=None):
        values = self.get_values()
        self.settings.set(f"algorithms.{self.name}", deepcopy(values))
        self.algorithm_thread.set_parameters(**values)
        self.algorithm_thread.start()

    def hideEvent(self, a0: QHideEvent):
        self.algorithm_thread.clean()

    def resizeEvent(self, event: QResizeEvent):
        if self.height() < self.form_widget.height():
            self.setMinimumWidth(self.form_widget.width() + 20)
        else:
            self.setMinimumWidth(self.form_widget.width() + 10)
        super().resizeEvent(event)

    def recursive_get_values(self):
        return self.form_widget.recursive_get_values()


class AlgorithmSettingsWidget(BaseAlgorithmSettingsWidget):
    def execute(self, exclude_mask=None):
        self.algorithm_thread.algorithm.set_image(self.settings.image)
        super().execute(exclude_mask)


class InteractiveAlgorithmSettingsWidget(BaseAlgorithmSettingsWidget):
    algorithm_thread: SegmentationThread

    def __init__(self, settings, name, algorithm: typing.Type[ROIExtractionAlgorithm], selector: typing.List[QWidget]):
        super().__init__(settings, name, algorithm)
        self.selector = selector
        self.algorithm_thread.finished.connect(self.enable_selector)
        self.algorithm_thread.started.connect(self.disable_selector)
        # noinspection PyUnresolvedReferences
        if hasattr(settings, "mask_changed"):
            settings.mask_changed.connect(self.change_mask)

    def value_updated(self):
        if not self.parent().interactive:
            return
        self.execute()

    def change_mask(self):
        if not self.isVisible():
            return
        self.algorithm_thread.algorithm.set_mask(self.settings.mask)

    def disable_selector(self):
        for el in self.selector:
            el.setDisabled(True)

    def enable_selector(self):
        for el in self.selector:
            el.setEnabled(True)

    def get_segmentation_profile(self) -> ROIExtractionProfile:
        return ROIExtractionProfile("", self.algorithm.get_name(), self.get_values())


class AlgorithmChooseBase(QWidget):
    finished = Signal()
    started = Signal()
    result = Signal(ROIExtractionResult)
    value_changed = Signal()
    progress_signal = Signal(str, int)
    algorithm_changed = Signal(str)

    def __init__(
        self, settings: BaseSettings, algorithms: typing.Dict[str, typing.Type[ROIExtractionAlgorithm]], parent=None
    ):
        super().__init__(parent)
        self.settings = settings
        self.algorithms = algorithms
        settings.algorithm_changed.connect(self.updated_algorithm)
        self.stack_layout = QStackedLayout()
        self.algorithm_choose = QComboBox()
        self.algorithm_dict: typing.Dict[str, BaseAlgorithmSettingsWidget] = {}
        self.algorithm_choose.currentTextChanged.connect(self.change_algorithm)
        self.add_widgets_to_algorithm()

        # self.setMinimumWidth(370)

        self.setContentsMargins(0, 0, 0, 0)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.algorithm_choose)
        layout.addLayout(self.stack_layout)
        self.setLayout(layout)

    @staticmethod
    def _algorithm_widget(settings, name, val) -> InteractiveAlgorithmSettingsWidget:
        return InteractiveAlgorithmSettingsWidget(settings, name, val, [])

    def add_widgets_to_algorithm(self):
        self.algorithm_choose.blockSignals(True)
        self.algorithm_choose.clear()
        for name, val in self.algorithms.items():
            self.algorithm_choose.addItem(name)
            widget = self._algorithm_widget(self.settings, name, val)
            self.algorithm_dict[name] = widget
            widget.algorithm_thread.execution_done.connect(self.result.emit)
            widget.algorithm_thread.finished.connect(self.finished.emit)
            widget.algorithm_thread.started.connect(self.started.emit)
            widget.algorithm_thread.progress_signal.connect(self.progress_signal.emit)
            widget.values_changed.connect(self.value_changed.emit)
            self.stack_layout.addWidget(widget)
        name = self.settings.get("current_algorithm", "")
        self.algorithm_choose.blockSignals(False)
        if name:
            self.algorithm_choose.setCurrentText(name)

    def reload(self, algorithms=None):
        if algorithms is not None:
            self.algorithms = algorithms
        for _ in range(self.stack_layout.count()):
            widget: InteractiveAlgorithmSettingsWidget = self.stack_layout.takeAt(0).widget()
            widget.algorithm_thread.execution_done.disconnect()
            widget.algorithm_thread.finished.disconnect()
            widget.algorithm_thread.started.disconnect()
            widget.algorithm_thread.progress_signal.disconnect()
            widget.values_changed.disconnect()
            widget.deleteLater()
        self.algorithm_dict = {}
        self.add_widgets_to_algorithm()

    def updated_algorithm(self):
        self.change_algorithm(
            self.settings.last_executed_algorithm,
            self.settings.get(f"algorithms.{self.settings.last_executed_algorithm}"),
        )

    def recursive_get_values(self):
        result = {key: widget.recursive_get_values() for key, widget in self.algorithm_dict.items()}

        self.settings.set("algorithm_widget_state", update(self.settings.get("algorithm_widget_state", {}), result))
        return result

    def change_algorithm(self, name, values: dict = None):
        self.settings.set("current_algorithm", name)
        widget = self.stack_layout.currentWidget()
        self.blockSignals(True)
        if name != widget.name:
            widget = self.algorithm_dict[name]
            self.stack_layout.setCurrentWidget(widget)
            widget.image_changed(self.settings.image)
            if hasattr(widget, "set_mask") and hasattr(self.settings, "mask"):
                widget.set_mask(self.settings.mask)
        elif values is None:
            self.blockSignals(False)
            return
        if values is not None:
            widget.set_values(values)
        self.algorithm_choose.setCurrentText(name)
        self.blockSignals(False)
        self.algorithm_changed.emit(name)

    def current_widget(self) -> InteractiveAlgorithmSettingsWidget:
        return self.stack_layout.currentWidget()

    def current_parameters(self) -> ROIExtractionProfile:
        widget = self.current_widget()
        return ROIExtractionProfile("", widget.name, widget.get_values())

    def get_info_text(self):
        return self.current_widget().algorithm_thread.get_info_text()


class AlgorithmChoose(AlgorithmChooseBase):
    def __init__(
        self, settings: BaseSettings, algorithms: typing.Dict[str, typing.Type[ROIExtractionAlgorithm]], parent=None
    ):
        super().__init__(settings, algorithms, parent)
        self.settings.image_changed.connect(self.image_changed)

    def image_changed(self):
        current_widget: InteractiveAlgorithmSettingsWidget = self.stack_layout.currentWidget()
        current_widget.image_changed(self.settings.image)
        if hasattr(self.settings, "mask") and hasattr(current_widget, "change_mask"):
            current_widget.change_mask()


# AbstractAlgorithmSettingsWidget.register(AlgorithmSettingsWidget)


def _value_get(self):
    return self.value


def _value_set(self, value):
    if isinstance(self, ComboBox) and issubclass(self.annotation, Layer) and isinstance(value, str):
        for el in self.choices:
            if el.name == value:
                self.value = el
                return
    self.value = value



import importlib
import os
import warnings
from functools import partial

from qtpy.QtCore import QSize, Qt, QThread
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QGridLayout, QMainWindow, QMessageBox, QProgressBar, QToolButton, QWidget

from PartSeg import ANALYSIS_NAME, APP_NAME, MASK_NAME
from PartSeg.common_backend import napari_get_settings
from PartSeg.common_backend.base_settings import BaseSettings, get_stylesheet, get_theme, napari_template
from PartSeg.common_backend.load_backup import import_config
from PartSeg.common_gui.main_window import BaseMainWindow
from PartSegCore import state_store
from PartSegData import icons_dir
from PartSegImage import TiffImageReader


class Prepare(QThread):
    def __init__(self, module):
        super().__init__()
        self.module = module
        self.result = None
        self.errors = []

    def run(self):
        if self.module != "":
            from .. import plugins

            plugins.register()
            main_window_module = importlib.import_module(self.module)
            main_window: BaseMainWindow = main_window_module.MainWindow
            settings: BaseSettings = main_window.get_setting_class()(main_window_module.CONFIG_FOLDER)
            self.errors = settings.load()
            reader = TiffImageReader()
            im = reader.read(main_window.initial_image_path)
            im.file_path = ""
            self.result = partial(main_window, settings=settings, initial_image=im)


class MainWindow(QMainWindow):
    def __init__(self, title):
        super().__init__()
        self.setWindowTitle(title)
        self.lib_path = ""
        self.final_title = ""
        analysis_icon = QIcon(os.path.join(icons_dir, "icon.png"))
        stack_icon = QIcon(os.path.join(icons_dir, "icon_stack.png"))
        self.analysis_button = QToolButton(self)
        self.analysis_button.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.analysis_button.setIcon(analysis_icon)
        # TODO use more general solution for text wrapping
        self.analysis_button.setText(ANALYSIS_NAME.replace(" ", "\n"))
        self.analysis_button.setIconSize(QSize(100, 100))
        self.mask_button = QToolButton(self)
        self.mask_button.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.mask_button.setIcon(stack_icon)
        self.mask_button.setText(MASK_NAME.replace(" ", "\n"))
        self.mask_button.setIconSize(QSize(100, 100))
        self.analysis_button.clicked.connect(self.launch_analysis)
        self.mask_button.clicked.connect(self.launch_mask)
        self.progress = QProgressBar()
        self.progress.setHidden(True)
        layout = QGridLayout()
        layout.addWidget(self.progress, 0, 0, 1, 2)
        layout.addWidget(self.analysis_button, 1, 1)
        layout.addWidget(self.mask_button, 1, 0)
        widget = QWidget()
        widget.setLayout(layout)
        self.setCentralWidget(widget)
        self.setWindowIcon(analysis_icon)
        self.prepare = None
        self.wind = None
        self._update_theme()

    def _update_theme(self):
        napari_settings = napari_get_settings(state_store.save_folder)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", FutureWarning)
            theme = get_theme(napari_settings.appearance.theme)
        # TODO understand qss overwrite mechanism
        self.setStyleSheet(napari_template(get_stylesheet(), **theme))

    def _launch_begin(self):
        self.progress.setVisible(True)
        self.progress.setRange(0, 0)
        self.analysis_button.setDisabled(True)
        self.mask_button.setDisabled(True)
        import_config()

    def launch_analysis(self):
        self._launch_begin()
        self._launch_analysis()
        self.prepare.start()

    def _launch_analysis(self):
        self.lib_path = "PartSeg._roi_analysis.main_window"
        self.final_title = f"{APP_NAME} {ANALYSIS_NAME}"
        self.prepare = Prepare(self.lib_path)
        self.prepare.finished.connect(self.launch)

    def launch_mask(self):
        self._launch_begin()
        self._launch_mask()
        self.prepare.start()

    def _launch_mask(self):
        self.lib_path = "PartSeg._roi_mask.main_window"
        self.final_title = f"{APP_NAME} {MASK_NAME}"
        self.prepare = Prepare(self.lib_path)
        self.prepare.finished.connect(self.launch)

    def window_shown(self):
        self.close()

    def launch(self):
        if self.prepare.result is None:
            self.close()
            return
        if self.prepare.errors:
            errors_message = QMessageBox()
            errors_message.setText("There are errors during start")
            errors_message.setInformativeText(
                "During load saved state some of data could not be load properly\n"
                "The files has prepared backup copies in  state directory (Help > State directory)"
            )
            errors_message.setStandardButtons(QMessageBox.Ok)
            text = "\n".join("File: " + x[0] + "\n" + str(x[1]) for x in self.prepare.errors)

            errors_message.setDetailedText(text)
            errors_message.exec_()
        wind = self.prepare.result(title=self.final_title, signal_fun=self.window_shown)
        wind.show()
        self.wind = wind



"""
This module contains non gui Qt based components
"""
import os.path
from typing import TYPE_CHECKING

import napari
import packaging.version

if TYPE_CHECKING:  # pragma: no cover
    from napari.settings import NapariSettings

try:
    from napari.settings import get_settings as _napari_get_settings
except ImportError:  # pragma: no cover
    try:
        from napari.utils.settings import get_settings as _napari_get_settings
    except ImportError:
        from napari.utils.settings import SETTINGS

        def _napari_get_settings(path=None):
            return SETTINGS


def napari_get_settings(path=None) -> "NapariSettings":
    if path is not None and packaging.version.parse(napari.__version__) >= packaging.version.parse("0.4.11"):
        path = os.path.join(path, "settings.yaml")

    try:
        return _napari_get_settings(path)
    except:  # noqa  # pylint: disable=W0702
        return _napari_get_settings()


__all__ = ("napari_get_settings",)



import importlib
import os
import sys
from importlib.abc import Loader, MetaPathFinder


class MyLoader(Loader):
    def module_repr(self, module):
        return repr(module)

    def load_module(self, fullname):
        old_name = fullname
        names = fullname.split(".")
        names[1] = "PartSegImage"
        fullname = ".".join(names[1:])
        module = importlib.import_module(fullname)
        sys.modules[old_name] = module
        return module


class MyImport(MetaPathFinder):
    def find_module(self, fullname, path=None):
        names = fullname.split(".")
        if len(names) >= 2 and names[0] == "PartSeg" and names[1] == "tiff_image":
            return MyLoader()


sys.meta_path.append(MyImport())

print(
    "[Warning] PartSeg.tiff_image module name is deprecated. It is renamed to PartSegImage"
    "To fail this import set environment variable 'NO_DEPRECATED' to 1",
    file=sys.stderr,
)

if "NO_DEPRECATED" in os.environ and os.environ["NO_DEPRECATED"] == "1":
    raise ImportError("PartSeg.tiff_image is deprecated and 'NO_DEPRECATED' is set to 1")



from ..json_hooks import ProfileEncoder, profile_hook
from .calculation_plan import CalculationPlan, CalculationTree
from .measurement_calculation import MeasurementProfile


class PartEncoder(ProfileEncoder):
    # pylint: disable=E0202
    def default(self, o):
        if isinstance(o, MeasurementProfile):
            return {"__MeasurementProfile__": True, **o.to_dict()}
        if isinstance(o, CalculationPlan):
            return {"__CalculationPlan__": True, "tree": o.execution_tree, "name": o.name}
        if isinstance(o, CalculationTree):
            return {"__CalculationTree__": True, "operation": o.operation, "children": o.children}
        return super().default(o)


def part_hook(dkt):
    if "__StatisticProfile__" in dkt:
        del dkt["__StatisticProfile__"]
        res = MeasurementProfile(**dkt)
        return res
    if "__MeasurementProfile__" in dkt:
        del dkt["__MeasurementProfile__"]
        res = MeasurementProfile(**dkt)
        return res
    if "__CalculationPlan__" in dkt:
        del dkt["__CalculationPlan__"]
        return CalculationPlan(**dkt)
    if "__CalculationTree__" in dkt:
        return CalculationTree(operation=dkt["operation"], children=dkt["children"])
    if (
        "__subtype__" in dkt
        and "statistic_profile" in dkt
        and dkt["__subtype__"]
        in (
            "PartSegCore.analysis.calculation_plan.MeasurementCalculate",
            "PartSeg.utils.analysis.calculation_plan.StatisticCalculate",
        )
    ):
        dkt["measurement_profile"] = dkt["statistic_profile"]
        del dkt["statistic_profile"]
    return profile_hook(dkt)



import typing
from concurrent.futures import ThreadPoolExecutor, as_completed

import numpy as np

from PartSegCore_compiled_backend.color_image_cython import color_grayscale, resolution

from .base_colors import BaseColormap

color_array_dict: typing.Dict[BaseColormap, np.ndarray] = {}
# TODO maybe replace with better cache structure


def create_color_map(colormap_definition: BaseColormap, power: float = 1.0) -> np.ndarray:
    """
    Calculate array with approximation of colormap used by :py:func`.color_image_fun` function.
    If first or last color do not have position 0 or 1 respectively then begin or end will be filled with given color

    Greyscale colormap
    ``res = create_color_map([ColorPosition(0, Color(0, 0, 0)), ColorPosition(1, Color(255, 255, 255))])``

    Black in first 25% of colormap then Greyscale from 25% to 75% and then white from 75% of colormap
    ``res = create_color_map([ColorPosition(0.25, Color(0, 0, 0)), ColorPosition(0.75, Color(255, 255, 255))])``

    :param colormap_definition: list defining base colormap
    :param power: power normalization of colormap. Will be ignored if points_and_colors.points be non empty
    :return: array of size (1024, 3)
    """
    # TODO Rethink
    colormap = np.zeros((1024, 3), dtype=np.uint8)
    bounds = colormap_definition.bounds()
    values = colormap_definition.color_values()
    if len(bounds) == 0:
        return colormap
    if len(bounds) == 1:
        colormap[:] = values[0]
        return colormap
    _bounds = [x ** power * ((resolution - 1) / resolution) for x in bounds]
    bounds = [_bounds[0]]
    _values = values
    values = [_values[0]]
    if len(bounds) < 10:
        for i, (x, y) in enumerate(zip(_bounds, _bounds[1:]), start=1):
            dist = (y - x) * resolution / 256
            if dist > 1:
                bounds.append(y - dist / resolution)
                values.append(_values[i])
            bounds.append(y)
            values.append(_values[i])
    values = np.array(values)
    points = list(colormap_definition.get_points())
    if not points:
        points = np.linspace(0, 1, resolution, endpoint=False)
    for i in range(3):
        colormap[:, i] = np.interp(points, bounds, values[:, i])
    return colormap


def color_bar_fun(bar: np.ndarray, colormap: typing.Union[BaseColormap, np.ndarray]):
    if isinstance(colormap, BaseColormap):
        if colormap not in color_array_dict:
            color_array_dict[colormap] = create_color_map(colormap)
        colormap = color_array_dict[colormap]
    if not isinstance(colormap, np.ndarray) or colormap.shape != (
        resolution,
        3,
    ):
        raise ValueError(f"colormap should be passed as numpy array with shape ({resolution},3)")
    min_val = np.min(bar)
    cords = ((bar - min_val) * ((colormap.shape[0] - 1) / (np.max(bar) - min_val))).astype(np.uint16)
    return colormap[cords]


def color_image_fun(
    image: np.ndarray,
    colors: typing.List[typing.Union[BaseColormap, np.ndarray]],
    min_max: typing.List[typing.Tuple[float, float]],
) -> np.ndarray:
    """
    Color given image layer.
    :param image: Single image layer (array of size (width, height, channels)
    :param colors: list of colormaps by name (from ``.color_maps`` dict) or array (its size ned to be size (1024, 3))
      array can be created by :py:func:`.create_color_map`
    :param min_max: bounds for each channel separately
    :return: colored image (array of size (width, height, 3) as RGB image
    """
    new_shape = image.shape[:-1] + (3,)

    result_images = []  # = np.zeros(new_shape, dtype=np.uint8)
    colored_channels = {}
    with ThreadPoolExecutor(max_workers=8) as executor:
        for i, colormap in enumerate(colors):
            if colormap is None:
                continue
            if isinstance(colormap, BaseColormap):
                if colormap not in color_array_dict:
                    color_array_dict[colormap] = create_color_map(colormap)
                colormap = color_array_dict[colormap]
            if not isinstance(colormap, np.ndarray) or not colormap.shape == (resolution, 3):
                raise ValueError(f"colormap should be passed as numpy array with shape ({resolution},3)")
            min_val, max_val = min_max[i]

            colored_channels[executor.submit(color_grayscale, colormap, image[..., i], min_val, max_val)] = i
    for res in as_completed(colored_channels):
        result_images.append(res.result())
    if len(result_images) > 0:
        if len(result_images) == 1:
            return result_images[0]
        # TODO use ColorMap additional information
        return np.max(result_images, axis=0)
    return np.zeros(new_shape, dtype=np.uint8)



from contextlib import suppress
from functools import partial

from qtpy.QtCore import Signal
from qtpy.QtWidgets import QCheckBox, QDialog, QHBoxLayout, QLabel, QPushButton, QSpinBox, QVBoxLayout, QWidget
from superqt import QEnumComboBox

from PartSegCore.image_operations import RadiusType
from PartSegCore.mask_create import MaskProperty
from PartSegCore.segmentation.algorithm_base import calculate_operation_radius

from ..common_backend.base_settings import BaseSettings, ImageSettings


def off_widget(widget: QWidget, combo_box: QEnumComboBox):
    widget.setDisabled(combo_box.currentEnum() == RadiusType.NO)


class MaskWidget(QWidget):
    values_changed = Signal()

    def __init__(self, settings: ImageSettings, parent=None):
        super().__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)
        self.settings = settings
        self.dilate_radius = QSpinBox()
        self.dilate_radius.setRange(-100, 100)
        # self.dilate_radius.setButtonSymbols(QAbstractSpinBox.NoButtons)
        self.dilate_radius.setSingleStep(1)
        self.dilate_radius.setDisabled(True)
        self.dilate_dim = QEnumComboBox(enum_class=RadiusType)
        self.dilate_dim.setToolTip("With minus radius mask will be eroded")
        # noinspection PyUnresolvedReferences
        self.dilate_dim.currentIndexChanged.connect(partial(off_widget, self.dilate_radius, self.dilate_dim))

        self.radius_information = QLabel()

        # noinspection PyUnresolvedReferences
        self.dilate_radius.valueChanged.connect(self.dilate_change)
        # noinspection PyUnresolvedReferences
        self.dilate_dim.currentIndexChanged.connect(self.dilate_change)

        self.fill_holes = QEnumComboBox(enum_class=RadiusType)
        self.max_hole_size = QSpinBox()
        self.max_hole_size.setRange(-1, 10000)
        self.max_hole_size.setValue(-1)
        self.max_hole_size.setSingleStep(100)
        self.max_hole_size.setDisabled(True)
        self.max_hole_size.setToolTip("Maximum size of holes to be closed. -1 means that all holes will be closed")
        # noinspection PyUnresolvedReferences
        self.fill_holes.currentIndexChanged.connect(partial(off_widget, self.max_hole_size, self.fill_holes))

        self.save_components = QCheckBox()
        self.clip_to_mask = QCheckBox()
        self.reversed_check = QCheckBox()

        # noinspection PyUnresolvedReferences
        self.dilate_radius.valueChanged.connect(self.values_changed.emit)
        # noinspection PyUnresolvedReferences
        self.dilate_dim.currentIndexChanged.connect(self.values_changed.emit)
        # noinspection PyUnresolvedReferences
        self.fill_holes.currentIndexChanged.connect(self.values_changed.emit)
        # noinspection PyUnresolvedReferences
        self.max_hole_size.valueChanged.connect(self.values_changed.emit)
        self.save_components.stateChanged.connect(self.values_changed.emit)
        self.clip_to_mask.stateChanged.connect(self.values_changed.emit)
        self.reversed_check.stateChanged.connect(self.values_changed.emit)

        layout = QVBoxLayout()
        layout1 = QHBoxLayout()
        layout1.addWidget(QLabel("Dilate mask:"))
        layout1.addWidget(self.dilate_dim)
        layout1.addWidget(QLabel("radius (in pix):"))
        layout1.addWidget(self.dilate_radius)
        layout.addLayout(layout1)
        layout2 = QHBoxLayout()
        layout2.addWidget(QLabel("Fill holes:"))
        layout2.addWidget(self.fill_holes)
        layout2.addWidget(QLabel("Max size:"))
        layout2.addWidget(self.max_hole_size)
        layout.addLayout(layout2)
        layout3 = QHBoxLayout()
        comp_lab = QLabel("Save components:")
        comp_lab.setToolTip(
            "save components information in mask. Dilation, " "holes filing will be done separately for each component"
        )
        self.save_components.setToolTip(comp_lab.toolTip())
        layout3.addWidget(comp_lab)
        layout3.addWidget(self.save_components)
        layout3.addStretch()
        clip_mask = QLabel("Clip to previous mask:")
        clip_mask.setToolTip("Useful dilated new mask")
        layout3.addWidget(clip_mask)
        layout3.addWidget(self.clip_to_mask)
        layout.addLayout(layout3)
        layout4 = QHBoxLayout()
        layout4.addWidget(QLabel("Reversed mask:"))
        layout4.addWidget(self.reversed_check)
        layout4.addStretch(1)
        layout.addLayout(layout4)
        self.setLayout(layout)
        self.dilate_change()

    def get_dilate_radius(self):
        radius = calculate_operation_radius(
            self.dilate_radius.value(), self.settings.image_spacing, self.dilate_dim.currentEnum()
        )
        if isinstance(radius, (list, tuple)):
            return [int(x + 0.5) for x in radius]
        return int(radius)

    def dilate_change(self):
        if self.dilate_radius.value() == 0 or self.dilate_dim.currentEnum() == RadiusType.NO:
            self.radius_information.setText("Dilation radius: 0")
        else:
            dilate_radius = self.get_dilate_radius()
            if isinstance(dilate_radius, list):
                self.radius_information.setText(f"Dilation radius: {dilate_radius[::-1]}")
            else:
                self.radius_information.setText(f"Dilation radius: {dilate_radius}")

    def get_mask_property(self):
        return MaskProperty(
            dilate=self.dilate_dim.currentEnum() if self.dilate_radius.value() != 0 else RadiusType.NO,
            dilate_radius=self.dilate_radius.value() if self.dilate_dim.currentEnum() != RadiusType.NO else 0,
            fill_holes=self.fill_holes.currentEnum() if self.max_hole_size.value() != 0 else RadiusType.NO,
            max_holes_size=self.max_hole_size.value() if self.fill_holes.currentEnum() != RadiusType.NO else 0,
            save_components=self.save_components.isChecked(),
            clip_to_mask=self.clip_to_mask.isChecked(),
            reversed_mask=self.reversed_check.isChecked(),
        )

    def set_mask_property(self, prop: MaskProperty):
        self.dilate_dim.setCurrentEnum(prop.dilate)
        self.dilate_radius.setValue(prop.dilate_radius)
        self.fill_holes.setCurrentEnum(prop.fill_holes)
        self.max_hole_size.setValue(prop.max_holes_size)
        self.save_components.setChecked(prop.save_components)
        self.clip_to_mask.setChecked(prop.clip_to_mask)
        self.reversed_check.setChecked(prop.reversed_mask)


class MaskDialogBase(QDialog):
    def __init__(self, settings: BaseSettings):
        super().__init__()
        self.setWindowTitle("Mask manager")
        self.settings = settings
        main_layout = QVBoxLayout()
        self.mask_widget = MaskWidget(settings, self)
        main_layout.addWidget(self.mask_widget)
        with suppress(KeyError):
            mask_property = self.settings.get("mask_manager.mask_property")
            self.mask_widget.set_mask_property(mask_property)

        self.reset_next_btn = QPushButton("Reset Next")
        self.reset_next_btn.clicked.connect(self.reset_next_fun)
        if settings.history_redo_size() == 0:
            self.reset_next_btn.setDisabled(True)
        self.set_next_btn = QPushButton("Set Next")
        if settings.history_redo_size() == 0:
            self.set_next_btn.setDisabled(True)
        self.set_next_btn.clicked.connect(self.set_next)
        self.cancel = QPushButton("Cancel", self)
        self.cancel.clicked.connect(self.close)
        self.prev_button = QPushButton(f"Previous mask ({settings.history_size()})", self)
        if settings.history_size() == 0:
            self.prev_button.setDisabled(True)
        self.next_button = QPushButton(f"Next mask ({settings.history_redo_size()})", self)
        if settings.history_redo_size() == 0:
            self.next_button.setText("Next mask (new)")
        self.next_button.clicked.connect(self.next_mask)
        self.prev_button.clicked.connect(self.prev_mask)
        op_layout = QHBoxLayout()
        op_layout.addWidget(self.mask_widget.radius_information)
        main_layout.addLayout(op_layout)
        button_layout = QHBoxLayout()
        button_layout.addWidget(self.cancel)
        button_layout.addWidget(self.set_next_btn)
        button_layout.addWidget(self.reset_next_btn)
        main_layout.addLayout(button_layout)
        button_layout = QHBoxLayout()
        button_layout.addWidget(self.prev_button)
        button_layout.addWidget(self.next_button)
        main_layout.addLayout(button_layout)
        self.setLayout(main_layout)
        if self.settings.history_redo_size():
            mask_prop: MaskProperty = self.settings.history_next_element().mask_property
            self.mask_widget.set_mask_property(mask_prop)
        self.mask_widget.values_changed.connect(self.values_changed)

    def set_next(self):
        if self.settings.history_redo_size():
            self.mask_widget.set_mask_property(self.settings.history_next_element().mask_property)

    def values_changed(self):
        if (
            self.settings.history_redo_size()
            and self.mask_widget.get_mask_property() == self.settings.history_next_element().mask_property
        ):
            self.next_button.setText(f"Next mask ({self.settings.history_redo_size()})")
        else:
            self.next_button.setText("Next mask (new)")

    def reset_next_fun(self):
        self.settings.history_redo_clean()
        self.next_button.setText("Next mask (new)")
        self.reset_next_btn.setDisabled(True)

    def next_mask(self):
        raise NotImplementedError()

    def prev_mask(self):
        raise NotImplementedError()



import typing

from napari.utils.colormaps.colormap import Colormap
from napari.utils.colormaps.colormap_utils import AVAILABLE_COLORMAPS

from PartSegCore.color_image.color_data import sitk_labels

Num = typing.Union[int, float]


class Color(typing.NamedTuple):
    """
    store color information

    :param red: red color value
    :param green: green color value
    :param blue: blue color value
    :param alpha: alpha value
    """

    red: float
    green: float
    blue: float
    alpha: float = 1


starting_colors = ["red", "green", "blue", "magenta", "inferno", "magma"]

default_colormap_dict = {name: AVAILABLE_COLORMAPS[name] for name in starting_colors}
default_colormap_dict.update(AVAILABLE_COLORMAPS)
default_colormap_dict.update(
    {
        k + "_reversed": Colormap(v.colors[::-1], controls=1 - v.controls[::-1])
        for k, v in AVAILABLE_COLORMAPS.items()
        if not k.endswith("_k")
    }
)


default_label_dict = {"default": sitk_labels}



import typing
from textwrap import indent

from PartSegCore.algorithm_describe_base import ROIExtractionProfile

from ..class_generator import BaseSerializableClass
from ..mask_create import MaskProperty


class SegmentationPipelineElement(BaseSerializableClass):
    segmentation: ROIExtractionProfile
    mask_property: MaskProperty

    def pretty_print(self, algorithm_dict):
        return (
            indent("Segmentation:\n" + self.segmentation.pretty_print(algorithm_dict).split("\n", 1)[1], "    ")
            + "\n\n"
            + indent(str(self.mask_property), "    ")
        )

    def __str__(self):
        return indent(str(self.segmentation), "    ") + "\n\n" + indent(str(self.mask_property), "    ")

    def __repr__(self):
        return (
            "SegmentationPipelineElement(segmentation="
            f"{repr(self.segmentation)},mask_property={repr(self.mask_property)})"
        )


class SegmentationPipeline(BaseSerializableClass):
    name: str
    segmentation: ROIExtractionProfile
    mask_history: typing.List[SegmentationPipelineElement]

    def pretty_print(self, algorithm_dict):
        return (
            (
                f"Segmentation pipeline name: {self.name}\n"
                + "\n––––––––––––––\n".join(x.pretty_print(algorithm_dict) for x in self.mask_history)
            )
            + "\n––––––––––––––\nLast segmentation:\n"
        ) + self.segmentation.pretty_print(algorithm_dict).split("\n", 1)[1]

    def __str__(self):
        return (
            (
                f"Segmentation pipeline name: {self.name}\n"
                + "\n––––––––––––––\n".join(str(x) for x in self.mask_history)
            )
            + "\n––––––––––––––\nLast segmentation\n"
        ) + str(self.segmentation)

    def __repr__(self):
        return (
            f"SegmentationPipeline(name={self.name},\nmask_history={self.mask_history},\n"
            f"segmentation={repr(self.segmentation)})"
        )



import os
from typing import Any, Optional

import numpy
from napari_plugin_engine import napari_hook_implementation

from PartSegCore.mask.io_functions import MaskProjectTuple, SaveROI
from PartSegCore.roi_info import ROIInfo
from PartSegImage.image import DEFAULT_SCALE_FACTOR


@napari_hook_implementation
def napari_write_labels(path: str, data: Any, meta: dict) -> Optional[str]:
    if not isinstance(data, numpy.ndarray):
        return
    ext = os.path.splitext(path)[1]
    if ext in SaveROI.get_extensions():
        project = MaskProjectTuple(file_path="", image=None, roi_info=ROIInfo(data))
        SaveROI.save(path, project, parameters={"spacing": numpy.divide(meta["scale"], DEFAULT_SCALE_FACTOR)[-3:]})
        return path



import os

import appdirs
import packaging.version

from PartSeg import APP_LAB, APP_NAME, __version__

report_errors = False
show_error_dialog = False
custom_plugin_load = False
check_for_updates = True
develop = False
save_suffix = ""
save_folder = os.path.join(
    appdirs.user_data_dir(APP_NAME, APP_LAB), str(packaging.version.parse(__version__).base_version)
)



import inspect
import typing
import warnings
from abc import ABC, abstractmethod
from collections import OrderedDict
from enum import Enum

from PartSegCore.channel_class import Channel


class AlgorithmDescribeNotFound(Exception):
    """
    When algorithm description not found
    """


class AlgorithmProperty:
    """
    This class is used to verbose describe algorithm parameters

    :param str name: name of parameter used in code
    :param str user_name: name presented to user in interface
    :param default_value: initial value which be used during interface generation
    :param str help_text: toll tip presented to user when keep mouse over widget
    :type value_type: type
    """

    def __init__(
        self,
        name: str,
        user_name: str,
        default_value: typing.Union[str, int, float, object],
        options_range=None,
        single_steep=None,
        possible_values=None,
        value_type=None,
        help_text="",
        per_dimension=False,
        **kwargs,
    ):
        if "property_type" in kwargs:
            warnings.warn("property_type is deprecated, use value_type instead", DeprecationWarning, stacklevel=2)
            value_type = kwargs["property_type"]
            del kwargs["property_type"]
        if len(kwargs) != 0:
            raise ValueError(", ".join(kwargs.keys()) + " are not expected")

        self.name = name
        self.user_name = user_name
        if isinstance(possible_values, list):
            self.value_type = list
        elif value_type is not None:
            self.value_type = value_type
        else:
            self.value_type = type(default_value)
        self.default_value = default_value
        self.range = options_range
        self.possible_values = possible_values
        self.single_step = single_steep
        self.help_text = help_text
        self.per_dimension = per_dimension
        if self.value_type is list and default_value not in possible_values:
            raise ValueError(f"default_value ({default_value}) should be one of possible values ({possible_values}).")

    def __repr__(self):
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}(name='{self.name}',"
            f" user_name='{self.user_name}', "
            + f"default_value={self.default_value}, type={self.value_type}, range={self.range},"
            f"possible_values={self.possible_values})"
        )


class AlgorithmDescribeBase(ABC):
    """
    This is abstract class for all algorithm exported to user interface.
    Based on get_name and get_fields methods the interface will be generated
    For each group of algorithm base abstract class will add additional methods
    """

    @classmethod
    def get_doc_from_fields(cls):
        resp = "{\n"
        for el in cls.get_fields():
            if isinstance(el, AlgorithmProperty):
                resp += f"  {el.name}: {el.value_type} - "
                if el.help_text:
                    resp += el.help_text
                resp += f"(default values: {el.default_value})\n"
        resp += "}\n"
        return resp

    @classmethod
    @abstractmethod
    def get_name(cls) -> str:
        """
        Algorithm name. It will be used during interface generating and in registering
        to proper :py:class:`PartSeg.PartSegCore.algorithm_describe_base.Register`.

        :return: name of algorithm
        """
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def get_fields(cls) -> typing.List[typing.Union[AlgorithmProperty, str]]:
        """
        This function return list of parameters needed by algorithm. It is used for generate form in User Interface

        :return: list of algorithm parameters and comments
        """
        raise NotImplementedError()

    @classmethod
    def get_fields_dict(cls) -> typing.Dict[str, AlgorithmProperty]:
        return {v.name: v for v in cls.get_fields() if isinstance(v, AlgorithmProperty)}

    @classmethod
    def get_default_values(cls):
        return {
            el.name: {
                "name": el.default_value,
                "values": el.possible_values[el.default_value].get_default_values(),
            }
            if issubclass(el.value_type, AlgorithmDescribeBase)
            else el.default_value
            for el in cls.get_fields()
            if isinstance(el, AlgorithmProperty)
        }


def is_static(fun):
    args = inspect.getfullargspec(fun).args
    if len(args) == 0:
        return True
    return args[0] != "self"


AlgorithmType = typing.TypeVar("AlgorithmType", bound=type(AlgorithmDescribeBase))


class Register(OrderedDict, typing.Generic[AlgorithmType]):
    """
    Dict used for register :class:`.AlgorithmDescribeBase` classes.
    All registers from `PartSeg.PartSegCore.register` are this
    :param class_methods: list of method which should be implemented as class method it will be checked during add
    as args or with :meth:`.Register.register`  method
    :param methods: list of method which should be instance method
    """

    def __init__(self, *args: AlgorithmType, class_methods=None, methods=None, suggested_base_class=None, **kwargs):
        """
        :param class_methods: list of method which should be class method
        :param methods: list of method which should be instance method
        :param kwargs: elements passed to OrderedDict constructor (may be initial elements). I suggest to not use this.
        """
        super().__init__(**kwargs)
        self.suggested_base_class = suggested_base_class
        self.class_methods = (
            list(class_methods) if class_methods else getattr(suggested_base_class, "need_class_method", [])
        )
        self.methods = list(methods) if methods else getattr(suggested_base_class, "need_method", [])
        for el in args:
            self.register(el)

    def values(self) -> typing.Iterable[AlgorithmType]:  # pylint: disable=W0235
        # noinspection PyTypeChecker
        return super().values()

    def __eq__(self, other):
        return (
            super().__eq__(other)
            and isinstance(other, Register)
            and self.class_methods == other.class_methods
            and self.methods == other.methods
            and self.suggested_base_class == other.suggested_base_class
        )

    def __getitem__(self, item) -> AlgorithmType:  # pylint: disable=W0235
        return super().__getitem__(item)

    def register(self, value: AlgorithmType, replace=False):
        """
        Function for registering :class:`.AlgorithmDescribeBase` based algorithms
        :param value: algorithm to register
        :param replace: replace existing algorithm, be patient with this
        """
        self.check_function(value, "get_name", True)
        try:
            name = value.get_name()
        except NotImplementedError:
            raise ValueError(f"Class {value} need to implement get_name class method")
        if name in self and not replace:
            raise ValueError(f"Object with this name: {name} already exist and register is not in replace mode")
        if not isinstance(name, str):
            raise ValueError(f"Function get_name of class {value} need return string not {type(name)}")
        self[name] = value

    @staticmethod
    def check_function(ob, function_name, is_class):
        fun = getattr(ob, function_name, None)
        if not is_class and not inspect.isfunction(fun):
            raise ValueError(f"Class {ob} need to define method {function_name}")
        if is_class and not inspect.ismethod(fun) and not is_static(fun):
            raise ValueError(f"Class {ob} need to define classmethod {function_name}")

    def __setitem__(self, key: str, value: AlgorithmType):
        if not issubclass(value, AlgorithmDescribeBase):
            raise ValueError(
                f"Class {value} need to inherit from " f"{AlgorithmDescribeBase.__module__}.AlgorithmDescribeBase"
            )
        self.check_function(value, "get_name", True)
        self.check_function(value, "get_fields", True)
        try:
            val = value.get_name()
        except NotImplementedError:
            raise ValueError(f"Method get_name of class {value} need to be implemented")
        if not isinstance(val, str):
            raise ValueError(f"Function get_name of class {value} need return string not {type(val)}")
        if key != val:
            raise ValueError("Object need to be registered under name returned by gey_name function")
        try:
            val = value.get_fields()
        except NotImplementedError:
            raise ValueError(f"Method get_fields of class {value} need to be implemented")
        if not isinstance(val, list):
            raise ValueError(f"Function get_fields of class {value} need return list not {type(val)}")
        for el in self.class_methods:
            self.check_function(value, el, True)
        for el in self.methods:
            self.check_function(value, el, False)

        super().__setitem__(key, value)

    def get_default(self) -> str:
        """
        Calculate default algorithm name for given dict.

        :return: name of algorithm
        """
        try:
            return next(iter(self.keys()))
        except StopIteration:
            raise ValueError("Register does not contain any algorithm.")


class ROIExtractionProfile:
    """

    :ivar str ~.name: name for segmentation profile
    :ivar str ~.algorithm: Name of algorithm
    :ivar dict ~.values: algorithm parameters
    """

    def __init__(self, name: str, algorithm: str, values: dict):
        self.name = name
        self.algorithm = algorithm
        self.values = values

    def pretty_print(self, algorithm_dict):
        try:
            algorithm = algorithm_dict[self.algorithm]
        except KeyError:
            return str(self)
        if self.name in {"", "Unknown"}:
            return (
                "ROI extraction profile\nAlgorithm: "
                + self.algorithm
                + "\n"
                + self._pretty_print(self.values, algorithm.get_fields_dict())
            )
        return (
            "ROI extraction profile name: "
            + self.name
            + "\nAlgorithm: "
            + self.algorithm
            + "\n"
            + self._pretty_print(self.values, algorithm.get_fields_dict())
        )

    @classmethod
    def _pretty_print(
        cls, values: typing.MutableMapping, translate_dict: typing.Dict[str, AlgorithmProperty], indent=0
    ):
        res = ""
        for k, v in values.items():
            if k not in translate_dict:
                if isinstance(v, typing.MutableMapping):
                    res += " " * indent + f"{k}: {cls._pretty_print(v, {}, indent + 2)}\n"
                else:
                    res += " " * indent + f"{k}: {v}\n"
                continue
            desc = translate_dict[k]
            res += " " * indent + desc.user_name + ": "
            if issubclass(desc.value_type, Channel):
                res += str(Channel(v))
            elif issubclass(desc.value_type, AlgorithmDescribeBase):
                res += desc.possible_values[v["name"]].get_name()
                if v["values"]:
                    res += "\n"
                    res += cls._pretty_print(v["values"], desc.possible_values[v["name"]].get_fields_dict(), indent + 2)
            elif isinstance(v, typing.MutableMapping):
                res += cls._pretty_print(v, {}, indent + 2)
            else:
                res += str(v)
            res += "\n"
        return res[:-1]

    @classmethod
    def print_dict(cls, dkt, indent=0, name: str = ""):
        if isinstance(dkt, Enum):
            return dkt.name
        if not isinstance(dkt, typing.MutableMapping):
            # FIXME update in future method of proper printing channel number
            if name.startswith("channel") and isinstance(dkt, int):
                return dkt + 1
            return dkt
        return "\n" + "\n".join(
            " " * indent + f"{k.replace('_', ' ')}: {cls.print_dict(v, indent + 2, k)}" for k, v in dkt.items()
        )

    def __str__(self):
        return (
            "ROI extraction profile name: "
            + self.name
            + "\nAlgorithm: "
            + self.algorithm
            + self.print_dict(self.values)
        )

    def __repr__(self):
        return f"ROIExtractionProfile(name={self.name}, algorithm={repr(self.algorithm)}, values={self.values})"

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.name == other.name
            and self.algorithm == other.algorithm
            and self.values == other.values
        )



from abc import ABC
from typing import Callable, List, Optional

from PartSegCore.algorithm_describe_base import AlgorithmDescribeBase
from PartSegImage import Image


class TransformBase(AlgorithmDescribeBase, ABC):
    @classmethod
    def transform(
        cls, image: Image, arguments: dict, callback_function: Optional[Callable[[str, int], None]] = None
    ) -> Image:
        raise NotImplementedError

    @classmethod
    def get_fields_per_dimension(cls, component_list: List[str]):
        raise NotImplementedError

    @classmethod
    def calculate_initial(cls, image: Image):
        raise NotImplementedError



import argparse
import getpass
import locale
import os
import platform
import sys
import zlib
from typing import Optional, Sequence

import numpy as np
import sentry_sdk
import sentry_sdk.serializer
import sentry_sdk.utils
from sentry_sdk.utils import safe_repr as _safe_repr

import PartSeg
from PartSeg.common_backend.except_hook import my_excepthook
from PartSegCore import state_store
from PartSegCore.utils import numpy_repr


def proper_suffix(val: str):
    """
    check if val contains only alphanumeric characters

    :raise argparse.ArgumentTypeError: on validation error
    """
    if len(val) > 0 and not val.isalnum():
        raise argparse.ArgumentTypeError(f"suffix '{val}' need to contains only alpha numeric characters")
    return val


def proper_path(val: str):
    """
    Check if `val` is proper path in current system

    :raise argparse.ArgumentTypeError: on validation error
    """
    if os.path.exists(val) and os.path.isdir(val):
        return val
    try:
        os.makedirs(val)
    except OSError:
        raise argparse.ArgumentTypeError(f" Path {val} is not a valid path in this system")
    return val


class CustomParser(argparse.ArgumentParser):
    """
    Argument parser with set of predefined flags:

    #. ``--no_report`` - disable error reporting, still showing dialog with information.
       Set :py:data:`.state_store.report_errors`.
    #. ``--no_dialog`` - disable error reporting and showing dialog. Exceptions will be printed on stderr.
       Set :py:data:`.state_store.show_error_dialog`.
    #. ``--no_update`` - disable check for updates on application startup
        Set :py:data:`.state_store.check_for_updates`.
    #. ``--save_suffix`` - add special suffix to directory contains saved state. designed to allow separate projects.
        Set :py:data:`.state_store.save_suffix`.
    #. ``--save_directory`` - set custom save directory. default can be previewed in `Help > State directory`
         from PartSeg main menu, Set :py:data:`.state_store.save_folder`.
    #. ``--inner_plugins`` - designed to hide part plugin from default start.
         Set :py:data:`.state_store.custom_plugin_load`.
    #. ``--develop`` -- for developer purpose. Allow to reload part of Program without restarting. May be unstable.
         Set :py:data:`.state_store.develop`. Base on this :py:class:`PartSeg.common_gui.advanced_tabs.AdvancedWindow`
         constructor add developer tab..

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_argument("--no_report", action="store_false", help="disable error reporting")
        self.add_argument("--no_dialog", action="store_false", help="disable error reporting and showing error dialog")
        self.add_argument("--no_update", action="store_false", help="disable check for updates")
        self.add_argument(
            "--save_suffix",
            "--ssuf",
            type=proper_suffix,
            default=[""],
            help="suffix for configuration_directory",
            nargs=1,
            metavar="suffix",
        )
        self.add_argument(
            "--save_directory",
            "--sdir",
            type=proper_path,
            default=[state_store.save_folder],
            help="path to custom configuration folder",
            nargs=1,
            metavar="path",
        )
        self.add_argument("--inner_plugins", action="store_true", help=argparse.SUPPRESS)
        self.add_argument("--develop", action="store_true", help=argparse.SUPPRESS)

    def parse_args(self, args: Optional[Sequence[str]] = None, namespace: Optional[argparse.Namespace] = None):
        """
        overload of :py:meth:`argparse.ArgumentParser.parse_args`. Set flags like described in class documentation.
        """
        args = super().parse_args(args, namespace)
        state_store.report_errors = args.no_report
        state_store.show_error_dialog = args.no_dialog
        state_store.custom_plugin_load = args.inner_plugins
        state_store.check_for_updates = args.no_update
        state_store.develop = args.develop
        state_store.save_suffix = args.save_suffix[0]
        state_store.save_folder = os.path.abspath(
            args.save_directory[0] + ("_" + state_store.save_suffix if state_store.save_suffix else "")
        )
        if args.no_report and args.no_dialog:
            _setup_sentry()
        sys.excepthook = my_excepthook
        locale.setlocale(locale.LC_NUMERIC, "")
        return args


def _setup_sentry():  # pragma: no cover
    sentry_sdk.utils.MAX_STRING_LENGTH = 10 ** 4
    sentry_sdk.serializer.safe_repr = safe_repr
    sentry_sdk.serializer.MAX_DATABAG_BREADTH = 100
    sentry_sdk.init(
        "https://d4118280b73d4ee3a0222d0b17637687@sentry.io/1309302",
        release=f"PartSeg@{PartSeg.__version__}",
    )
    with sentry_sdk.configure_scope() as scope:
        scope.set_user(
            {"name": getpass.getuser(), "id": zlib.adler32((getpass.getuser() + "#" + platform.node()).encode())}
        )


def safe_repr(val):
    if isinstance(val, np.ndarray):
        return numpy_repr(val)
    return _safe_repr(val)



from ..algorithm_describe_base import Register
from ..segmentation.restartable_segmentation_algorithms import final_algorithm_list

analysis_algorithm_dict = Register(
    *final_algorithm_list,
    class_methods=["support_time", "support_z"],
    methods=["set_image", "set_mask", "get_info_text", "calculation_run"],
)
"""Register for segmentation method designed for separate specific areas."""



import sys
from abc import ABC
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Union

import numpy as np
from sympy import Symbol, symbols

from PartSegImage.image import Spacing

from ..algorithm_describe_base import AlgorithmDescribeBase, AlgorithmDescribeNotFound, AlgorithmProperty
from ..channel_class import Channel
from ..class_generator import BaseSerializableClass, enum_register
from ..universal_const import Units


class PerComponent(Enum):
    No = 1
    Yes = 2
    Mean = 3

    def __str__(self):
        return self.name.replace("_", " ")


class AreaType(Enum):
    ROI = 1
    Mask = 2
    Mask_without_ROI = 3

    def __str__(self):
        return self.name.replace("_", " ")


enum_register.register_class(AreaType)
enum_register.register_class(PerComponent)


class Leaf(BaseSerializableClass):
    """
    Class for describe calculation of basic measurement

    :ivar str name: node name of method used to calculate
    :ivar dict dict: additional parameters of method
    :ivar float power: power to be applied to result of calculation methods
    :ivar AreaType area: which type of ROI should be used for calculation
    :ivar PerComponent per_component: if value should be calculated per component or for whole roi set
    :ivar Channel channel: probably not used TODO Check
    """

    # noinspection PyMissingConstructor,PyShadowingBuiltins, PyUnusedLocal
    # pylint: disable=W0104,W0221,W0622,W0231
    def __init__(
        self,
        name: str,
        dict: Dict = None,
        power: float = 1.0,
        area: Optional[AreaType] = None,
        per_component: Optional[PerComponent] = None,
        channel: Optional[Channel] = None,
    ):
        ...

    name: str
    dict: Dict = {}
    power: float = 1.0
    area: Optional[AreaType] = None
    per_component: Optional[PerComponent] = None
    channel: Optional[Channel] = None

    def get_channel_num(self, measurement_dict: Dict[str, "MeasurementMethodBase"]) -> Set[Channel]:
        """
        Get set with number of channels needed for calculate this measurement

        :param measurement_dict: dict with all measurementh method.
        :return: set of channels num
        """
        resp = set()
        if self.channel is not None and self.channel >= 0:
            resp.add(self.channel)
        try:
            measurement_method = measurement_dict[self.name]
            for el in measurement_method.get_fields():
                if isinstance(el, str):
                    continue
                if issubclass(el.value_type, Channel) and el.name in self.dict:
                    resp.add(self.dict[el.name])
        except KeyError:
            raise AlgorithmDescribeNotFound(self.name)
        return resp

    def _parameters_string(self, measurement_dict: Dict[str, "MeasurementMethodBase"]) -> str:
        if len(self.dict) == 0 and self.channel is None:
            return ""
        arr = []
        if self.channel is not None and self.channel >= 0:
            arr.append(f"channel={self.channel+1}")
        if self.name in measurement_dict:
            measurement_method = measurement_dict[self.name]
            fields_dict = measurement_method.get_fields_dict()
            for k, v in self.dict.items():
                arr.append(f"{fields_dict[k].user_name}={v}")
        else:
            for k, v in self.dict.items():
                arr.append(f"{k.replace('_', ' ')}={v}")
        return "[" + ", ".join(arr) + "]"

    def _plugin_info(self, measurement_dict: Dict[str, "MeasurementMethodBase"]) -> str:
        if self.name not in measurement_dict:
            return ""
        measurement_method = measurement_dict[self.name]
        if (
            hasattr(measurement_method, "__module__")
            and measurement_method.__module__.split(".", 1)[0] != "PartSegCore"
        ):
            if getattr(sys, "frozen", False):
                return f"[{measurement_method.__module__.split('.', 2)[1]}] "
            return f"[{measurement_method.__module__.split('.', 1)[0]}] "
        return ""

    def pretty_print(self, measurement_dict: Dict[str, "MeasurementMethodBase"]) -> str:
        resp = self.name
        if self.area is not None:
            resp = str(self.area) + " " + resp
        resp = self._plugin_info(measurement_dict) + resp
        if self.per_component is not None:
            if self.per_component == PerComponent.Yes:
                resp += " per component "
            elif self.per_component == PerComponent.Mean:
                resp += " mean component "
        resp += self._parameters_string(measurement_dict)
        if self.power != 1.0:
            resp += f" to the power {self.power}"
        return resp

    def __str__(self):  # pragma: no cover
        return self.pretty_print({})

    def get_unit(self, ndim) -> Symbol:
        from PartSegCore.analysis import MEASUREMENT_DICT

        method = MEASUREMENT_DICT[self.name]
        if self.power != 1:
            return method.get_units(ndim) ** self.power
        return method.get_units(ndim)

    def is_per_component(self) -> bool:
        return self.per_component == PerComponent.Yes


def replace(self, **kwargs) -> Leaf:
    for key in list(kwargs.keys()):
        if key == "power":
            continue
        if not hasattr(self, key):
            raise ValueError(f"Unknown parameter {key}")
        if getattr(self, key) is not None and (key != "dict" or self.dict):
            del kwargs[key]

    dkt = self.asdict()
    dkt.update(kwargs)
    return Leaf(**dkt)


Leaf.replace_ = replace


class Node(BaseSerializableClass):
    left: Union["Node", Leaf]
    op: str
    right: Union["Node", Leaf]

    # noinspection PyMissingConstructor, PyUnusedLocal
    # pylint: disable=W0104,W0231
    def __init__(self, left: Union["Node", Leaf], op: str, right: Union["Node", Leaf]):
        ...

    def get_channel_num(self, measurement_dict: Dict[str, "MeasurementMethodBase"]) -> Set[Channel]:
        return self.left.get_channel_num(measurement_dict) | self.right.get_channel_num(measurement_dict)

    def __str__(self):  # pragma: no cover
        left_text = "(" + str(self.left) + ")" if isinstance(self.left, Node) else str(self.left)
        right_text = "(" + str(self.right) + ")" if isinstance(self.right, Node) else str(self.right)
        return left_text + self.op + right_text

    def pretty_print(self, measurement_dict: Dict[str, "MeasurementMethodBase"]) -> str:  # pragma: no cover
        left_text = (
            "(" + self.left.pretty_print(measurement_dict) + ")"
            if isinstance(self.left, Node)
            else self.left.pretty_print(measurement_dict)
        )
        right_text = (
            "(" + self.right.pretty_print(measurement_dict) + ")"
            if isinstance(self.right, Node)
            else self.right.pretty_print(measurement_dict)
        )
        return left_text + self.op + right_text

    def get_unit(self, ndim) -> Symbol:
        if self.op == "/":
            return self.left.get_unit(ndim) / self.right.get_unit(ndim)
        raise ValueError(f"Unknown operator '{self.op}'")

    def is_per_component(self) -> bool:
        return self.left.is_per_component() or self.right.is_per_component()


class MeasurementEntry(BaseSerializableClass):
    # noinspection PyUnusedLocal
    # noinspection PyMissingConstructor
    __old_names__ = "StatisticEntry"

    # noinspection PyMissingConstructor,PyUnusedLocal
    # pylint: disable=W0104,W0231
    def __init__(self, name: str, calculation_tree: Union[Node, Leaf]):
        ...

    name: str
    calculation_tree: Union[Node, Leaf]

    def get_unit(self, unit: Units, ndim) -> str:
        return str(self.calculation_tree.get_unit(ndim)).format(str(unit))

    def get_channel_num(self, measurement_dict: Dict[str, "MeasurementMethodBase"]) -> Set[Channel]:
        return self.calculation_tree.get_channel_num(measurement_dict)


class MeasurementMethodBase(AlgorithmDescribeBase, ABC):
    """
    This is base class For all measurement calculation classes
    based on text_info[0] the measurement name wil be generated, based_on text_info[1] the description is generated
    """

    text_info = "", ""

    need_class_method = [
        "get_description",
        "is_component",
        "calculate_property",
        "get_starting_leaf",
        "get_units",
        "need_channel",
    ]

    @classmethod
    def get_name(cls) -> str:
        return str(cls.get_starting_leaf().name)

    @classmethod
    def get_description(cls) -> str:
        """Measurement long description"""
        if isinstance(cls.text_info, str):
            return ""
        return cls.text_info[1]

    @classmethod
    def is_component(cls) -> bool:
        """Return information if Need information about components"""
        return False

    @classmethod
    def get_fields(cls) -> List[Union[str, AlgorithmProperty]]:
        """Additional fields needed by algorithm. like radius of dilation"""
        return []

    @staticmethod
    def calculate_property(
        # image: Image,
        channel: np.ndarray,
        roi: np.ndarray,
        mask: np.ndarray,
        voxel_size: Spacing,
        result_scalar: float,
        roi_alternative: Dict[str, np.ndarray],
        roi_annotation: Dict[int, Any],
        **kwargs,
    ):
        """
        Main function for calculating measurement

        :param channel: main channel selected for measurement
        :param channel_{i}: for channel requested using :py:meth:`get_fields`
            ``AlgorithmProperty("channel", "Channel", 0, value_type=Channel)``
        :param area_array: array representing current area returned by :py:meth:`area_type`
        :param roi: array representing roi
        :param mask: array representing mask (upper level roi)
        :param voxel_size: size of single voxel in meters
        :param result_scalar: scalar to get proper units in result
        :param roi_alternative: dict with alternative roi representation (for plugin specific mapping)
        :param roi_annotation: dict with roi annotations (for plugin specific mapping)

        List incomplete.
        """
        raise NotImplementedError()

    @classmethod
    def get_starting_leaf(cls) -> Leaf:
        """This leaf is putted on default list"""
        return Leaf(cls._display_name())

    @classmethod
    def _display_name(cls):
        if isinstance(cls.text_info, str):
            return cls.text_info
        return cls.text_info[0]

    @classmethod
    def get_units(cls, ndim) -> symbols:
        """Return units for measurement. They are shown to user"""
        raise NotImplementedError()

    @classmethod
    def need_channel(cls):
        """if need image data"""
        return False

    @staticmethod
    def area_type(area: AreaType):
        """Map chosen area type to proper area type. Allow to correct Area type."""
        return area

    @staticmethod
    def need_full_data():
        return False



import typing
from abc import ABC
from enum import Enum

import numpy as np

from ..algorithm_describe_base import AlgorithmDescribeBase, AlgorithmProperty, Register
from ..class_generator import enum_register
from ..image_operations import bilateral, gaussian, median
from .algorithm_base import calculate_operation_radius as _calculate_operation_radius


class DimensionType(Enum):
    Layer = 1
    Stack = 2

    def __str__(self):
        return self.name.replace("_", " ")


try:
    # noinspection PyUnresolvedReferences,PyUnboundLocalVariable
    reloading
except NameError:
    reloading = False  # means the module is being imported
    enum_register.register_class(DimensionType, old_name="GaussType")


class NoiseFilteringBase(AlgorithmDescribeBase, ABC):
    """Base class for noise filtering operations"""

    @classmethod
    def noise_filter(cls, channel: np.ndarray, spacing: typing.Iterable[float], arguments: dict) -> np.ndarray:
        """
        This function need be overloaded in implementation

        :param channel: single channel ad 2d or 3d array
        :param spacing: image spacing
        :param arguments: additional arguments defined by :py:meth:`get_fields`
        :return: channel array with removed noise
        """
        raise NotImplementedError()


class NoneNoiseFiltering(NoiseFilteringBase):
    @classmethod
    def get_name(cls):
        return "None"

    @classmethod
    def get_fields(cls):
        return []

    @classmethod
    def noise_filter(cls, channel: np.ndarray, spacing: typing.Iterable[float], arguments: dict):
        return channel


class GaussNoiseFiltering(NoiseFilteringBase):
    @classmethod
    def get_name(cls):
        return "Gauss"

    @classmethod
    def get_fields(cls):
        return [
            AlgorithmProperty("dimension_type", "Gauss type", DimensionType.Layer),
            AlgorithmProperty("radius", "Gauss radius", 1.0, value_type=float),
        ]

    @classmethod
    def noise_filter(cls, channel: np.ndarray, spacing: typing.Iterable[float], arguments: dict):
        gauss_radius = calculate_operation_radius(arguments["radius"], spacing, arguments["dimension_type"])
        layer = arguments["dimension_type"] == DimensionType.Layer
        return gaussian(channel, gauss_radius, layer=layer)


class BilateralNoiseFiltering(NoiseFilteringBase):
    @classmethod
    def get_name(cls):
        return "Bilateral"

    @classmethod
    def get_fields(cls):
        return [
            AlgorithmProperty("dimension_type", "Gauss type", DimensionType.Layer),
            AlgorithmProperty("radius", "Gauss radius", 1.0, value_type=float),
        ]

    @classmethod
    def noise_filter(cls, channel: np.ndarray, spacing: typing.Iterable[float], arguments: dict):
        gauss_radius = calculate_operation_radius(arguments["radius"], spacing, arguments["dimension_type"])
        layer = arguments["dimension_type"] == DimensionType.Layer
        return bilateral(channel, max(gauss_radius), layer=layer)


def calculate_operation_radius(radius, spacing, gauss_type):
    res = _calculate_operation_radius(radius, spacing, gauss_type)
    if res == radius:
        return [radius for _ in spacing]
    return res


class MedianNoiseFiltering(NoiseFilteringBase):
    @classmethod
    def get_name(cls):
        return "Median"

    @classmethod
    def get_fields(cls):
        return [
            AlgorithmProperty("dimension_type", "Median type", DimensionType.Layer),
            AlgorithmProperty("radius", "Median radius", 1, value_type=int),
        ]

    @classmethod
    def noise_filter(cls, channel: np.ndarray, spacing: typing.Iterable[float], arguments: dict):
        gauss_radius = calculate_operation_radius(arguments["radius"], spacing, arguments["dimension_type"])
        layer = arguments["dimension_type"] == DimensionType.Layer
        gauss_radius = [int(x) for x in gauss_radius]
        return median(channel, gauss_radius, layer=layer)


noise_filtering_dict = Register(
    NoneNoiseFiltering,
    GaussNoiseFiltering,
    MedianNoiseFiltering,
    BilateralNoiseFiltering,
    class_methods=["noise_filter"],
)



import os
from typing import Any, Optional

import numpy
from napari_plugin_engine import napari_hook_implementation

from PartSegImage import Image, ImageWriter
from PartSegImage.image import DEFAULT_SCALE_FACTOR


@napari_hook_implementation
def napari_write_labels(path: str, data: Any, meta: dict) -> Optional[str]:
    ext = os.path.splitext(path)[1]
    if not isinstance(data, numpy.ndarray) or ext not in {".tiff", ".tif", ".TIFF", ".TIF"}:
        return
    scale_shift = min(data.ndim, 3)
    image = Image(
        data,
        numpy.divide(meta["scale"], DEFAULT_SCALE_FACTOR)[-scale_shift:],
        axes_order="TZXY"[-data.ndim :],
        channel_names=[meta["name"]],
        shift=numpy.divide(meta["translate"], DEFAULT_SCALE_FACTOR)[-scale_shift:],
        name="ROI",
    )
    ImageWriter.save(image, path)
    return path


@napari_hook_implementation
def napari_write_image(path: str, data: Any, meta: dict) -> Optional[str]:
    ext = os.path.splitext(path)[1]
    if not isinstance(data, numpy.ndarray) or ext not in {".tiff", ".tif", ".TIFF", ".TIF"}:
        return
    scale_shift = min(data.ndim, 3)
    axes = "TZXY"
    channel_names = [meta["name"]]
    if data.shape[-1] < 6:
        axes += "C"
        scale_shift -= 1
        channel_names = [f'{meta["name"]} {i}' for i in range(1, data.shape[-1] + 1)]
    image = Image(
        data,
        numpy.divide(meta["scale"], DEFAULT_SCALE_FACTOR)[-scale_shift:],
        axes_order=axes[-data.ndim :],
        channel_names=channel_names,
        shift=numpy.divide(meta["translate"], DEFAULT_SCALE_FACTOR)[-scale_shift:],
        name="Image",
    )
    ImageWriter.save(image, path)
    return path
